"""
report for strategy benchmark with tre
"""
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Optional, Tuple, Dict

import qis as qis
from qis import TimePeriod, PerfParams, BenchmarkReturnsQuantilesRegime
from qis.portfolio.multi_portfolio_data import MultiPortfolioData
from qis.portfolio.risk.factor_model import LinearModel
from qis.portfolio.reports.config import PERF_PARAMS, regime_classifier


def weights_tracking_error_report_by_ac_subac(multi_portfolio_data: MultiPortfolioData,
                                              strategy_idx: int = 0,
                                              benchmark_idx: int = 1,
                                              ac_group_data: pd.Series = None,
                                              ac_group_order: List[str] = None,
                                              sub_ac_group_data: pd.Series = None,
                                              sub_ac_group_order: List[str] = None,
                                              turnover_groups: pd.Series = None,
                                              turnover_order: List[str] = None,
                                              risk_model: LinearModel = None,
                                              time_period: TimePeriod = None,
                                              perf_params: PerfParams = PERF_PARAMS,
                                              regime_classifier: BenchmarkReturnsQuantilesRegime = BenchmarkReturnsQuantilesRegime(),
                                              add_benchmarks_to_navs: bool = True,
                                              tre_max_clip: Optional[float] = None,
                                              figsize: Tuple[float, float] = (11.7, 8.3),
                                              var_format: str = '{:.1%}',
                                              add_titles: bool = True,
                                              **kwargs
                                              ) -> Tuple[Dict[str, plt.Figure], Dict[str, pd.DataFrame]]:
    """
    weights and tracking error report relative to benchnmark using ac and sub_as groups
    """
    regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]
    benchmark_price = multi_portfolio_data.benchmark_prices[regime_benchmark]

    figs: Dict[str, plt.Figure] = {}
    dfs: Dict[str, pd.DataFrame] = {}

    with (sns.axes_style('darkgrid')):
        # navs + ra table
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['navs'] = fig
        if add_titles:
            title = f"Cumulative performance with background colors using bear/normal/bull "
            f"regimes of {regime_benchmark} {regime_classifier.freq}-returns"
        else:
            title = None
        multi_portfolio_data.plot_nav(regime_benchmark=regime_benchmark,
                                      time_period=time_period,
                                      perf_params=perf_params,
                                      regime_classifier=regime_classifier,
                                      add_benchmarks_to_navs=add_benchmarks_to_navs,
                                      title=title,
                                      ax=ax,
                                      **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['ra_table'] = fig
        dfs['nav'] = multi_portfolio_data.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        multi_portfolio_data.plot_ra_perf_table(benchmark_price=benchmark_price,
                                                                add_benchmarks_to_navs=add_benchmarks_to_navs,
                                                                perf_params=perf_params,
                                                                time_period=time_period,
                                                                add_turnover=True,
                                                                ax=ax,
                                                                **kwargs)
        # ra perf table without strings
        ra_perf_table = multi_portfolio_data.plot_ra_perf_table(benchmark_price=benchmark_price,
                                                                add_benchmarks_to_navs=add_benchmarks_to_navs,
                                                                perf_params=perf_params,
                                                                time_period=time_period,
                                                                add_turnover=True,
                                                                is_convert_to_str=False,
                                                                ax=ax,
                                                                **kwargs)
        dfs['ra_perf_table'] = ra_perf_table

        # strategy weights
        strategy_data = multi_portfolio_data.portfolio_datas[strategy_idx]
        strategy_ticker = strategy_data.ticker
        weight_kwargs = dict(is_grouped=True, time_period=time_period, add_total=False, is_input_weights=True)
        strategy_exposures_ac = strategy_data.get_weights(group_data=ac_group_data, group_order=ac_group_order,
                                                          **weight_kwargs)
        strategy_exposures_subac = strategy_data.get_weights(group_data=sub_ac_group_data, group_order=sub_ac_group_order,
                                                             **weight_kwargs)

        # benchmark weights
        benchmark_data = multi_portfolio_data.portfolio_datas[benchmark_idx]
        benchmark_ticker = benchmark_data.ticker
        benchmark_exposures_ac = benchmark_data.get_weights(group_data=ac_group_data, group_order=ac_group_order,
                                                            **weight_kwargs)
        benchmark_exposures_subac = benchmark_data.get_weights(group_data=sub_ac_group_data, group_order=sub_ac_group_order,
                                                               **weight_kwargs)

        # plot strategy and benchmark weights by ac
        kwargs = qis.update_kwargs(kwargs, dict(strategy_ticker=f"(B) {strategy_ticker}",
                                                benchmark_ticker=f"(A) {benchmark_ticker}"))
        fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
        if add_titles:
            qis.set_suptitle(fig, title=f"Time series of weights by asset classes")
        figs['strategy_benchmark_weights_stack'] = fig
        plot_exposures_strategy_vs_benchmark_stack(strategy_exposures=strategy_exposures_ac,
                                                   benchmark_exposures=benchmark_exposures_ac,
                                                   axs=axs,
                                                   var_format=var_format,
                                                   **kwargs)

        # boxplot by subac
        fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
        if add_titles:
            qis.set_suptitle(fig, title=f"Boxplot of weights")
        figs['strategy_benchmark_weights_box'] = fig
        plot_exposures_strategy_vs_benchmark_boxplot(strategy_exposures=strategy_exposures_ac,
                                                     benchmark_exposures=benchmark_exposures_ac,
                                                     ax=axs[0],
                                                     ylabel='Weights',
                                                     title='(A) Weights by asset classes',
                                                     hue_var_name='Asset Class',
                                                     var_format=var_format,
                                                     allow_negative=True,
                                                     **kwargs)
        plot_exposures_strategy_vs_benchmark_boxplot(strategy_exposures=strategy_exposures_subac,
                                                     benchmark_exposures=benchmark_exposures_subac,
                                                     ax=axs[1],
                                                     ylabel='Weights',
                                                     title='(B) Weights by sub-asset classes',
                                                     hue_var_name='Sub-Asset Class',
                                                     var_format=var_format,
                                                     **kwargs)
        # portfolio vol
        strategy_ex_anti_vol = strategy_data.compute_ex_anti_portfolio_vol_implied_by_covar(
            covar_dict=multi_portfolio_data.covar_dict)
        benchmark_ex_anti_vol = benchmark_data.compute_ex_anti_portfolio_vol_implied_by_covar(
            covar_dict=multi_portfolio_data.covar_dict)

        ex_anti_vols = pd.concat([strategy_ex_anti_vol, benchmark_ex_anti_vol], axis=1)
        dfs['ex_anti_vols'] = ex_anti_vols
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        if add_titles:
            qis.set_suptitle(fig, title=f"Ex-anti portfolio volatility")
        figs['ex_anti_vols'] = fig
        qis.plot_time_series(df=ex_anti_vols,
                             var_format='{:.2%}',
                             ax=ax,
                             **kwargs)
        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                    index=ex_anti_vols.index, regime_classifier=regime_classifier)

        # risk contributions
        rc_kwargs = dict(covar_dict=multi_portfolio_data.covar_dict, freq='QE', normalise=True, time_period=time_period)
        strategy_risk_contributions_ac = strategy_data.compute_risk_contributions_implied_by_covar(
            group_data=ac_group_data,
            group_order=ac_group_order,
            **rc_kwargs)
        strategy_risk_contributions_subac = strategy_data.compute_risk_contributions_implied_by_covar(
            group_data=sub_ac_group_data,
            group_order=sub_ac_group_order,
            **rc_kwargs)

        benchmark_risk_contributions_ac = benchmark_data.compute_risk_contributions_implied_by_covar(
            group_data=ac_group_data,
            group_order=ac_group_order,
            **rc_kwargs)
        benchmark_risk_contributions_subac = benchmark_data.compute_risk_contributions_implied_by_covar(
            group_data=sub_ac_group_data,
            group_order=sub_ac_group_order,
            **rc_kwargs)

        # stack for ac
        fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
        if add_titles:
            qis.set_suptitle(fig, title=f"Time Series of risk contributions by asset classes")
        figs['time_series_risk_contrib'] = fig
        plot_exposures_strategy_vs_benchmark_stack(strategy_exposures=strategy_risk_contributions_ac,
                                                   benchmark_exposures=benchmark_risk_contributions_ac,
                                                   axs=axs,
                                                   var_format=var_format,
                                                   **kwargs)

        # box plots for subac
        fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
        if add_titles:
            qis.set_suptitle(fig, title=f"Boxplot of risk contributions")
        figs['risk_contributions_boxplot'] = fig
        plot_exposures_strategy_vs_benchmark_boxplot(
            strategy_exposures=strategy_risk_contributions_ac,
            benchmark_exposures=benchmark_risk_contributions_ac,
            ax=axs[0],
            title='(A) Risk contributions by asset classes',
            hue_var_name='Asset Class',
            ylabel='Risk contributions',
            var_format=var_format,
            allow_negative=True,
            **kwargs)
        plot_exposures_strategy_vs_benchmark_boxplot(
            strategy_exposures=strategy_risk_contributions_subac,
            benchmark_exposures=benchmark_risk_contributions_subac,
            ax=axs[1],
            title='(B) Risk contributions by sub-asset classes',
            hue_var_name='Sub-Asset Class',
            ylabel='Risk contributions',
            var_format=var_format,
            allow_negative=True,
            **kwargs)

        # brinson by asset class
        totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return = \
            multi_portfolio_data.compute_brinson_attribution(strategy_idx=strategy_idx,
                                                             benchmark_idx=benchmark_idx,
                                                             time_period=time_period,
                                                             group_data=ac_group_data,
                                                             group_order=ac_group_order,
                                                             freq=None,
                                                             total_column='Total Sum',
                                                             is_exclude_interaction_term=True)
        figs['brinson_table_ac'] = qis.plot_brinson_totals_table(totals_table=totals_table, **kwargs)
        dfs['brinson_table_ac'] = totals_table

        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['brinson_total_time_series'] = fig
        if add_titles:
            title = 'Active total return'
        else:
            title = None
        qis.plot_time_series(df=active_total.cumsum(axis=0),
                             title=title,
                             legend_stats=qis.LegendStats.LAST,
                             var_format='{:.0%}',
                             ax=ax, **kwargs)
        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                    index=active_total.index, regime_classifier=regime_classifier)
        dfs['brinson_active_total'] = active_total.cumsum(axis=0)

        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['brinson_grouped_allocation_return'] = fig
        qis.plot_time_series(df=grouped_allocation_return.cumsum(axis=0),
                             title='Grouped allocation return',
                             legend_stats=qis.LegendStats.LAST,
                             var_format='{:.0%}',
                             ax=ax, **kwargs)
        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                    index=grouped_allocation_return.index, regime_classifier=regime_classifier)

        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['brinson_grouped_selection_return'] = fig
        qis.plot_time_series(df=grouped_selection_return.cumsum(axis=0),
                             title='Grouped selection return',
                             legend_stats=qis.LegendStats.LAST,
                             var_format='{:.0%}',
                             ax=ax, **kwargs)
        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                    index=grouped_selection_return.index, regime_classifier=regime_classifier)

        # brinson by sub-asset class
        totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return = \
            multi_portfolio_data.compute_brinson_attribution(strategy_idx=strategy_idx,
                                                             benchmark_idx=benchmark_idx,
                                                             time_period=time_period,
                                                             group_data=sub_ac_group_data,
                                                             group_order=sub_ac_group_order,
                                                             freq=None,
                                                             total_column='Total Sum',
                                                             is_exclude_interaction_term=True)
        figs['brinson_table_subac'] = qis.plot_brinson_totals_table(totals_table=totals_table, **kwargs)
        dfs['brinson_table_subac'] = totals_table

        # tracking error
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['tre_time_series'] = fig
        if add_titles:
            title = 'Tracking Error'
        else:
            title = None
        multi_portfolio_data.plot_tre_time_series(strategy_idx=strategy_idx,
                                                  benchmark_idx=benchmark_idx,
                                                  regime_benchmark=regime_benchmark,
                                                  regime_classifier=regime_classifier,
                                                  title=title,
                                                  ax=ax,
                                                  time_period=time_period,
                                                  **kwargs)

        # group tracking error
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['tre_group_time_series'] = fig
        if add_titles:
            title = 'Asset Class Tracking Error'
        else:
            title = None
        multi_portfolio_data.plot_tre_time_series(strategy_idx=strategy_idx,
                                                  benchmark_idx=benchmark_idx,
                                                  is_grouped=True,
                                                  group_data=ac_group_data,
                                                  group_order=ac_group_order,
                                                  regime_benchmark=regime_benchmark,
                                                  regime_classifier=regime_classifier,
                                                  tre_max_clip=tre_max_clip,
                                                  title=title,
                                                  ax=ax,
                                                  time_period=time_period,
                                                  **kwargs)
        dfs['ac_tracking_error'] = multi_portfolio_data.compute_tracking_error_implied_by_covar(strategy_idx=strategy_idx,
                                                                                                benchmark_idx=benchmark_idx,
                                                                                                is_grouped=True,
                                                                                                group_data=ac_group_data,
                                                                                                group_order=ac_group_order,
                                                                                                total_column='Total')

        # turnover
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['joint_turnover'] = fig
        multi_portfolio_data.plot_turnover(ax=ax,
                                           time_period=time_period,
                                           regime_benchmark=regime_benchmark,
                                           regime_classifier=regime_classifier,
                                           **kwargs)
        if not add_titles:
            ax.title.set_visible(False)

        # group turnover
        fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
        figs['group_turnover'] = fig
        multi_portfolio_data.portfolio_datas[strategy_idx].plot_turnover(ax=ax,
                                                                         time_period=time_period,
                                                                         regime_benchmark=regime_benchmark,
                                                                         regime_classifier=regime_classifier,
                                                                         is_grouped=True,
                                                                         group_data=turnover_groups,
                                                                         group_order=turnover_order,
                                                                         add_total=False,
                                                                         **kwargs)
        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                    index=grouped_selection_return.index, regime_classifier=regime_classifier)
        if not add_titles:
            ax.title.set_visible(False)

        dfs['ac_turnover'] = multi_portfolio_data.portfolio_datas[strategy_idx].get_turnover(is_agg=False,
                                                                                             is_grouped=True,
                                                                                             group_data=turnover_groups,
                                                                                             group_order=turnover_order,
                                                                                             time_period=time_period,
                                                                                             add_total=False,
                                                                                             **kwargs)

        # pdf of returns
        freqs = dict(Monthly='ME', Quarterly='QE', Annual='YE')
        fig, axs = plt.subplots(1, len(freqs.keys()), figsize=figsize, tight_layout=True)
        figs['returns_pdfs'] = fig
        navs = multi_portfolio_data.get_navs(time_period=time_period, add_benchmarks_to_navs=False)
        for idx, (key, freq) in enumerate(freqs.items()):
            returns = qis.to_returns(prices=navs, freq=freq, drop_first=True)
            if len(returns.index) > 3:
                qis.plot_histogram(df=returns,
                                   xvar_format='{:.0%}',
                                   add_bar_at_peak=True,
                                   desc_table_type=qis.DescTableType.NONE,
                                   title=f"({qis.idx_to_alphabet(idx+1)}) {key} Returns",
                                   xlabel='return',
                                   ax=axs[idx])
            #if not add_titles:
            #    ax.title.set_visible(False)

        # outputs with risk model
        if risk_model is not None:

            # factor - level
            out_dict = risk_model.compute_active_factor_risk(portfolio_weights=multi_portfolio_data.portfolio_datas[strategy_idx].get_weights(),
                                                             benchmark_weights=multi_portfolio_data.portfolio_datas[benchmark_idx].get_weights())

            # strategy factor betas
            strategy_factor_betas = out_dict['portfolio_exposures']
            # strategy_factor_betas = risk_model.compute_agg_factor_exposures(weights=multi_portfolio_data.portfolio_datas[strategy_idx].get_weights())
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['strategy_factor_betas'] = fig
            qis.plot_time_series(df=strategy_factor_betas,
                                 title=f"{strategy_ticker} Factor Beta Exposures",
                                 var_format='{:,.2f}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=strategy_factor_betas.index, regime_classifier=regime_classifier)
            # benchmark factor betas
            benchmark_factor_betas = out_dict['benchmark_exposures']
            # benchmark_factor_betas = risk_model.compute_agg_factor_exposures(weights=multi_portfolio_data.portfolio_datas[benchmark_idx].get_weights())
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['benchmark_factor_betas'] = fig
            qis.plot_time_series(df=benchmark_factor_betas,
                                 title=f"{benchmark_ticker} Factor Beta Exposures",
                                 var_format='{:,.2f}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=benchmark_factor_betas.index, regime_classifier=regime_classifier)

            # active exposure
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['active_exposure'] = fig
            qis.plot_stack(df=out_dict['active_exposures'],
                           legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                           title=f"{strategy_ticker} vs {benchmark_ticker} active exposure",
                           var_format='{:,.2f}',
                           ax=ax)
            # active factor risk
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['factor_risk_contributions_rc'] = fig
            qis.plot_stack(df=out_dict['factor_risk_contributions_rc'],
                           legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                           title=f"{strategy_ticker} vs {benchmark_ticker} active risk contribution %",
                           var_format='{:,.2%}',
                           ax=ax)

            # strategy attribution
            portfolio_returns = qis.to_returns(prices=multi_portfolio_data.portfolio_datas[strategy_idx].get_portfolio_nav().reindex(
                index=strategy_factor_betas.index).ffill(), is_first_zero=True)
            attributions = qis.compute_benchmarks_beta_attribution_from_returns(portfolio_returns=portfolio_returns,
                                                                                benchmark_returns=risk_model.x,
                                                                                portfolio_benchmark_betas=strategy_factor_betas,
                                                                                total_name='Total')
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['strategy_factor_attributions'] = fig
            qis.plot_time_series(df=attributions.cumsum(axis=0),
                                 title=f"{strategy_ticker} Factor Attribution",
                                 legend_stats=qis.LegendStats.LAST,
                                 var_format='{:,.1%}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=attributions.index, regime_classifier=regime_classifier)

            # benchmark attribution
            portfolio_returns = qis.to_returns(prices=multi_portfolio_data.portfolio_datas[benchmark_idx].get_portfolio_nav().reindex(
                index=benchmark_factor_betas.index).ffill(), is_first_zero=True)
            attributions = qis.compute_benchmarks_beta_attribution_from_returns(portfolio_returns=portfolio_returns,
                                                                                benchmark_returns=risk_model.x,
                                                                                portfolio_benchmark_betas=benchmark_factor_betas,
                                                                                total_name='Total')
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['benchmark_factor_attributions'] = fig
            qis.plot_time_series(df=attributions.cumsum(axis=0),
                                 title=f"{benchmark_ticker} Factor Attribution",
                                 legend_stats=qis.LegendStats.LAST,
                                 var_format='{:,.1%}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=attributions.index, regime_classifier=regime_classifier)

            # strategy risk attribution
            factor_rcs_ratios, strategy_factor_risk_contrib_idio, factor_risk_contrib, strategy_portfolio_var = \
                risk_model.compute_factor_risk_contribution(weights=multi_portfolio_data.portfolio_datas[strategy_idx].get_weights())
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['strategy_factor_risk_cb'] = fig
            qis.plot_stack(df=strategy_factor_risk_contrib_idio,
                           use_bar_plot=True,
                           title=f"{strategy_ticker} relative factor risk contribution",
                           var_format='{:,.2%}',
                           ax=ax)

            # benchmark attribution
            factor_rcs_ratios, benchmark_factor_risk_contrib_idio, factor_risk_contrib, benchmark_portfolio_var = \
                risk_model.compute_factor_risk_contribution(weights=multi_portfolio_data.portfolio_datas[benchmark_idx].get_weights())
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['benchmark_factor_risk_cb'] = fig
            qis.plot_stack(df=benchmark_factor_risk_contrib_idio,
                           use_bar_plot=True,
                           title=f"{benchmark_ticker} relative factor risk contribution",
                           var_format='{:,.2%}',
                           ax=ax)

            # joint risk contribs
            kwargs = qis.update_kwargs(kwargs, dict(strategy_ticker=f"(B) {strategy_ticker}",
                                                    benchmark_ticker=f"(A) {benchmark_ticker}"))
            fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
            if add_titles:
                qis.set_suptitle(fig, title=f"Relative risk contributions")
            figs['strategy_benchmark_risk_contributions'] = fig
            plot_exposures_strategy_vs_benchmark_stack(strategy_exposures=strategy_factor_risk_contrib_idio,
                                                       benchmark_exposures=benchmark_factor_risk_contrib_idio,
                                                       axs=axs,
                                                       var_format=var_format,
                                                       **kwargs)

            # portfolio vars
            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['strategy_portfolio_vars'] = fig
            qis.plot_time_series(df=np.sqrt(strategy_portfolio_var),
                                 title=f"{strategy_ticker} Portfolio sqrt(Vars)",
                                 var_format='{:,.2%}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=strategy_portfolio_var.index, regime_classifier=regime_classifier)

            fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
            figs['benchmark_portfolio_vars'] = fig
            qis.plot_time_series(df=np.sqrt(benchmark_portfolio_var),
                                 title=f"{benchmark_ticker} Portfolio sqrt(Vars)",
                                 var_format='{:,.2%}',
                                 ax=ax)
            if regime_benchmark is not None:
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark,
                                                        index=benchmark_portfolio_var.index, regime_classifier=regime_classifier)

    return figs, dfs


def plot_exposures_long_short_groups(exposures_short: pd.DataFrame,
                                     exposures_long: pd.DataFrame,
                                     axs: List[plt.Subplot],
                                     ylabel: str = 'weights',
                                     var_format: str = '{:.1%}',
                                     hue_var_name: str = 'asset class',
                                     **kwargs
                                     ) -> None:
    qis.plot_stack(df=exposures_short,
                   use_bar_plot=True,
                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                   var_format=var_format,
                   colors=qis.get_n_sns_colors(n=len(exposures_short.columns)),
                   ax=axs[0],
                   **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.01), ncols=1,
                                                    framealpha=0.9)))

    qis.df_boxplot_by_columns(df=exposures_long,
                              hue_var_name=hue_var_name,
                              y_var_name=ylabel,
                              ylabel=ylabel,
                              showmedians=True,
                              add_y_median_labels=False,
                              yvar_format=var_format,
                              x_rotation=90,
                              colors=qis.get_n_sns_colors(n=len(exposures_long.columns)),
                              y_limits=(0.0, None),
                              ax=axs[1],
                              **kwargs)


def plot_exposures_strategy_vs_benchmark_stack(strategy_exposures: pd.DataFrame,
                                               benchmark_exposures: pd.DataFrame,
                                               axs: List[plt.Subplot],
                                               var_format: str = '{:.1%}',
                                               strategy_ticker: str = 'TAA',
                                               benchmark_ticker: str = 'SAA',
                                               **kwargs
                                               ) -> None:
    qis.plot_stack(df=benchmark_exposures,
                   use_bar_plot=True,
                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                   var_format=var_format,
                   colors=qis.get_n_sns_colors(n=len(benchmark_exposures.columns)),
                   title=benchmark_ticker,
                   ax=axs[0],
                   **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.01), ncols=1,
                                                    framealpha=0.9)))
    qis.plot_stack(df=strategy_exposures,
                   use_bar_plot=True,
                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                   var_format=var_format,
                   colors=qis.get_n_sns_colors(n=len(strategy_exposures.columns)),
                   title=strategy_ticker,
                   ax=axs[1],
                   **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.01), ncols=1,
                                                    framealpha=0.9)))


def plot_exposures_strategy_vs_benchmark_boxplot(strategy_exposures: pd.DataFrame,
                                                 benchmark_exposures: pd.DataFrame,
                                                 ax: plt.Subplot,
                                                 ylabel: str = 'weights',
                                                 var_format: str = '{:.1%}',
                                                 hue_var_name: str = 'asset class',
                                                 strategy_ticker: str = 'TAA',
                                                 benchmark_ticker: str = 'SAA',
                                                 allow_negative: bool = False,
                                                 title: str = '',
                                                 **kwargs
                                                 ) -> None:
    dfs = {benchmark_ticker: benchmark_exposures, strategy_ticker: strategy_exposures}
    if allow_negative:
        y_limits = None
    else:
        y_limits = (0.0, None)
    qis.df_dict_boxplot_by_columns(dfs=dfs,
                                   hue_var_name=hue_var_name,
                                   y_var_name=ylabel,
                                   ylabel=ylabel,
                                   showmedians=True,
                                   add_y_median_labels=False,
                                   yvar_format=var_format,
                                   x_rotation=90,
                                   title=title,
                                   # colors=qis.get_n_sns_colors(n=len(exposures_long.columns)),
                                   y_limits=y_limits,
                                   ax=ax,
                                   **kwargs)



"""
common python structures operations
"""
import itertools
import pandas as pd
from collections.abc import Iterable
from enum import Enum
from typing import Dict, Tuple, List, Any, NamedTuple, Union, Optional


def list_intersection(list_check: Union[List[Any], pd.Index],
                      list_sample: Union[List[Any], pd.Index]
                      ) -> List[Any]:
    """Find the intersection of elements between two lists or pandas Index objects.

    Returns elements from list_sample that are also present in list_check,
    maintaining the order from list_sample.

    Args:
        list_check (Union[List[Any], pd.Index]): The reference list/Index to check
            against for membership.
        list_sample (Union[List[Any], pd.Index]): The list/Index to filter based
            on elements present in list_check.

    Returns:
        List[Any]: A list containing elements from list_sample that exist in
            list_check. Returns empty list if no intersection found.

    Example:
        >>> list_intersection([1, 2, 3], [2, 3, 4, 5])
        [2, 3]
    """
    list_out = list(filter(lambda x: x in list_check, list_sample))
    if list_out is None:
        list_out = []
    return list_out


def list_diff(list_check: Union[List[Any], pd.Index],
              list_sample: Union[List[Any], pd.Index]
              ) -> List[Any]:
    """Find elements in list_sample that are not present in list_check.

    Returns the difference between two lists, maintaining the order from
    list_sample for elements not found in list_check.

    Args:
        list_check (List[Any]): The reference list to check against for exclusion.
        list_sample (List[Any]): The list to filter, excluding elements that
            exist in list_check.

    Returns:
        List[Any]: A list containing elements from list_sample that do not
            exist in list_check. Returns empty list if all elements are found
            in list_check.

    Example:
        >>> list_diff([1, 2, 3], [2, 3, 4, 5])
        [4, 5]
    """
    list_out = list(filter(lambda x: x not in list_check, list_sample))
    if list_out is None:
        list_out = []
    return list_out


def merge_lists_unique(list1: Union[List[Any], pd.Index],
                       list2: Union[List[Any], pd.Index]
                       ) -> List[Any]:
    """Merge two lists while preserving all unique elements from both lists.

    Combines two lists by first finding their intersection, then appending
    unique elements from each list. The result contains all unique elements
    from both input lists without duplicates.

    Args:
        list1 (List[Any]): The first list to merge.
        list2 (List[Any]): The second list to merge.

    Returns:
        List[Any]: A merged list containing all unique elements from both
            input lists. Order is: intersection elements first, followed by
            unique elements from list1, then unique elements from list2.

    Example:
        >>> merge_lists_unique([1, 2, 3], [2, 3, 4, 5])
        [2, 3, 1, 4, 5]
    """
    list_out = list_intersection(list_check=list2, list_sample=list1)
    list_diffs1 = list_diff(list_check=list_out, list_sample=list1)
    list_diffs2 = list_diff(list_check=list_out, list_sample=list2)
    if len(list_diffs1) > 0:
        list_out.extend(list_diffs1)
    if len(list_diffs2) > 0:
        list_out.extend(list_diffs2)
    return list_out


def assert_list_subset(large_list: Union[List[Any], pd.Index],  # larger list
                       list_sample: Union[List[Any], pd.Index],  # smaller list,
                       is_stop: bool = True,
                       message: str = ''
                       ) -> bool:
    outlyers = list_diff(list_check=large_list, list_sample=list_sample)
    output = True
    if len(outlyers) > 0:
        if is_stop:
            raise ValueError(f"{message}\nnot found {outlyers}")
        else:
            print(f"in assert_list_subset:\n{message}\n{outlyers}")
            output = False
    return output


def list_to_unique_and_dub(lsdata: Union[List[Any], pd.Index]) -> Tuple[List, List]:
    """
    find unique and dublicates in list
    """
    unique = set()
    dublicated = []  # can count occurencies if needed
    for x in lsdata:
        if x not in unique:
            unique.add(x)
        else:
            dublicated.append(x)
    unique = list(unique)
    return unique, dublicated


def assert_list_unique(lsdata: Union[List[str], pd.Index], name: str = '') -> None:
    unique, duplicated = list_to_unique_and_dub(lsdata=lsdata)
    if len(duplicated) > 0:
        raise ValueError(f"list {name} has duplicated elements = {duplicated}")


def move_item_to_first(lsdata: Union[List[Any], pd.Index], item: Any) -> List[Any]:
    out_list = lsdata.copy()
    out_list.remove(item)
    out_list = [item]+out_list
    return out_list


def flatten_dict_tuples(dict_tuples: Dict[str, Union[Any, NamedTuple]]) -> Dict[str, Any]:
    data = {}
    for k, v in dict_tuples.items():
        if isinstance(v, tuple):
            data.update(v._asdict())  # this will create a dict from named tuple
        else:
            data[k] = v
    return data


def split_dict(d: Dict) -> Tuple[Dict, Dict]:
    """
    split dictionary into 2 parts
    """
    n = len(d) // 2
    i = iter(d.items())

    d1 = dict(itertools.islice(i, n))   # grab first n items
    d2 = dict(i)                        # grab the rest

    return d1, d2


def flatten(items: Iterable) -> Any:
    """
    flatten list/items from any nested iterable
    """
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x


def to_flat_list(items: Iterable) -> List[Any]:
    if isinstance(items, Iterable):
        flat_list = [item for item in flatten(items)]
    else:
        flat_list = [items]
    return flat_list


def update_kwargs(kwargs: Dict[Any, Any],
                  new_kwargs: Optional[Dict[Any, Any]]
                  ) -> Dict[Any, Any]:
    """
    update kwargs with optional kwargs dicts
    """
    local_kwargs = kwargs.copy()
    if new_kwargs is not None and not len(new_kwargs) == 0:
        local_kwargs.update(new_kwargs)
    return local_kwargs


def separate_number_from_string(string: str) -> List[str]:
    """
    given 'A3' get 3
    """
    previous_character = string[0]
    groups = []
    newword = string[0]
    for x, i in enumerate(string[1:]):
        if i.isalpha() and previous_character.isalpha():
            newword += i
        elif i.isnumeric() and previous_character.isnumeric():
            newword += i
        else:
            groups.append(newword)
            newword = i
        previous_character = i
        if x == len(string) - 2:
            groups.append(newword)
            newword = ''

    return groups


class LocalTests(Enum):
    FLATTEN = 1
    LIST = 2
    LIST_INTERSECTION = 3
    MERGE = 4
    LIST_DIFF = 5
    STRINGS = 6


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.FLATTEN:
        items = [[1, 2], [[3]], 4]
        flat_items = flatten(items)
        [print(item) for item in flat_items]
        print(to_flat_list(items))

    elif local_test == LocalTests.LIST:
        rows_edge_lines = list(itertools.accumulate(10 * [5]))
        print(rows_edge_lines)

    elif local_test == LocalTests.LIST_INTERSECTION:
        list2 = ['EQ', 'HUI', 'Metals']
        list1 = ['EQ', 'BD', 'STIR', 'FX', 'Energies', 'Metals', 'Ags']
        groups = list_intersection(list_check=list1, list_sample=list2)
        print('groups1')
        print(groups)
        groups = list_intersection(list_check=list2, list_sample=list1)
        print('groups2')
        print(groups)

    elif local_test == LocalTests.MERGE:
        list2 = ['EQ', 'HUI']
        list1 = ['EQ', 'BD', 'STIR', 'FX', 'Energies', 'Metals', 'Ags']
        groups = merge_lists_unique(list1=list1, list2=list2)
        print('groups1')
        print(groups)
        groups = merge_lists_unique(list1=list2, list2=list1)
        print('groups2')
        print(groups)

    elif local_test == LocalTests.LIST_DIFF:
        list2 = ['EQ', 'HUI']
        list1 = ['EQ', 'BD', 'STIR', 'FX', 'Energies', 'Metals', 'Ags']
        this = list_diff(list_check=list1, list_sample=list2)
        print(this)
        this = list_diff(list_check=list2, list_sample=list1)
        print(this)

    elif local_test == LocalTests.STRINGS:
        string = '123me45you0000me7+33.3'
        this = separate_number_from_string(string)
        print(this)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.LIST_INTERSECTION)



"""
boxplot
"""
# packages
import warnings
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Union, List, Dict, Optional, Tuple
from enum import Enum

# qis
import qis.utils.df_cut as dfc
import qis.utils.df_melt as dfm
import qis.plots.utils as put


def plot_box(df: Union[pd.Series, pd.DataFrame],
             x: str = None,
             y: str = None,
             xlabel: Optional[Union[str, bool]] = True,
             ylabel: Optional[Union[str, bool]] = True,
             hue: Optional[str] = None,
             hue_order: List[str] = None,
             xvar_format: str = '{:.2f}',
             yvar_format: str = '{:.2f}',
             colors: Union[List[str], List[Tuple[float, float, float]]] = None,
             title: str = None,
             meanline: bool = False,
             showfliers: bool = False,  # outliers
             showmeans: bool = False,
             showmedians: bool = False,
             add_y_mean_labels: bool = False,
             add_xy_mean_labels: bool = False,
             add_xy_median_labels: bool = False,
             add_y_median_labels: bool = False,
             x_rotation: Optional[int] = 0,
             legend_loc: Optional[str] = 'upper right',
             fontsize: int = 10,
             linewidth: float = 1.0,
             y_lines: List[Dict] = None,
             labels: List[str] = None,
             original_index: Union[pd.DatetimeIndex, pd.Index] = None,
             x_date_freq: Union[str, None] = 'YE',
             date_format: str = '%b-%y',
             num_obs_for_ci: np.ndarray = None,
             add_zero_line: bool = False,
             add_mean_std_bound: bool = False,
             add_autocorr_std_bound: bool = False,
             continuous_x_col: str = None,
             y_limits: Tuple[Optional[float], Optional[float]] = None,
             whis: Optional[float] = 1.5,  # sns default
             add_hue_to_legend_title: bool = True,
             ax: plt.Subplot = None,
             **kwargs
             ) -> Optional[plt.Figure]:
    """
    plot boxplot of df[[x, y]]
    original_index use for melted df in term of original index
    """
    if x is None and y is None:
        if len(df.columns) == 2:
            x = df.columns[0]
            y = df.columns[1]
        else:
            raise ValueError(f"x and y must be defined for more than one columns")

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is not None:
        palette = colors
    else:
        if hue is None:
            palette = put.get_data_group_colors(df=df, x=x, y=y)
        else:
            palette = put.get_n_colors(n=len(df[x].unique()))

    sns.boxplot(x=x, y=y, data=df,
                hue=hue or x,
                hue_order=hue_order,
                palette=palette,
                linewidth=linewidth,
                meanline=meanline,
                showmeans=showmeans,
                showfliers=showfliers,
                medianprops={'visible': showmedians, 'color': 'black'},
                meanprops={'visible': showmeans or meanline, 'color': 'black'},
                whis=whis,
                ax=ax)

    if num_obs_for_ci is not None:
        se = 1.96 / np.sqrt(num_obs_for_ci)
        mean = np.zeros_like(se)
        xx = np.arange(0, len(num_obs_for_ci))
        ax.plot(xx, -se+mean, color='red', lw=linewidth, linestyle='--')
        ax.plot(xx, mean, color='red', lw=linewidth, linestyle='-')
        ax.plot(xx, se + mean, color='red', lw=linewidth, linestyle='--')

    elif add_mean_std_bound:
        mean = np.nanmean(df[y].to_numpy())
        stdev = np.nanstd(df[y].to_numpy())
        se = 1.96 * stdev / np.sqrt(len(df.index))
        y_lines = [dict(x=-se+mean, color='blue', linestyle='--'),
                   dict(x=mean, color='blue', linestyle='-'),
                   dict(x=se+mean, color='blue', linestyle='--')]
    elif add_autocorr_std_bound:
        se = 1.96 * 2.0 / np.sqrt(len(df.index))
        y_lines = [dict(x=-se, color='blue', linestyle='--'),
                   dict(x=0, color='blue', linestyle='-'),
                   dict(x=se, color='blue', linestyle='--')]

    elif add_zero_line:
        y_lines = [dict(x=0, color='black', linestyle='-')]

    if y_lines is not None:
        for y_line in y_lines:
            ax.axhline(y=y_line['x'], color=y_line['color'], linestyle=y_line['linestyle'], lw = 1)

    # x ticks
    if original_index is not None:
        dates = pd.Series(data=original_index, index=original_index)
        re_indexed_data, datalables = put.map_dates_index_to_str(data=dates,
                                                                 x_date_freq=x_date_freq,
                                                                 date_format=date_format)

        current_ticks = ax.get_xticks()
        xticks = list(np.linspace(current_ticks[0], current_ticks[-1], len(datalables)))
        put.set_ax_tick_labels(ax=ax, x_labels=datalables, x_rotation=x_rotation, xticks=xticks, fontsize=fontsize)

    else:

        # ax.set_xticklabels(ax.get_xticklabels(), fontsize=fontsize, rotation=x_rotation, minor=False)
        put.set_ax_tick_params(ax=ax, fontsize=fontsize)
        put.set_ax_tick_labels(ax=ax, x_rotation=x_rotation, fontsize=fontsize)

    if len(ax.get_yticks()) > 0:
        ax.set_yticks(ax.get_yticks())
        ax.set_yticklabels([yvar_format.format(x) for x in ax.get_yticks()],
                           fontsize=fontsize, minor=False)

    if add_y_mean_labels or add_xy_mean_labels:
        if hue is not None:
            df_by_x = df.sort_values(hue).groupby(hue, sort=False)
            x_col = x
        else:
            df_by_x = df.sort_values(x).groupby(x, sort=False)
            if continuous_x_col is None:
                raise ValueError(f"continuous_x_col must be given")
            x_col = continuous_x_col

        for x_idx, (key, g_data) in enumerate(df_by_x):
            y_val = g_data[y].mean()
            x_val = g_data[x_col].mean()
            if add_y_mean_labels:
                ax.text(x_idx, y_val, f"{yvar_format.format(y_val)}\n",
                        ha='center', va='center',
                        color='black', fontsize=fontsize)
            else:
                ax.text(x_idx, y_val, f"avg:\ny={yvar_format.format(y_val)}\nx={xvar_format.format(x_val)}",
                        ha='center', va='center',
                        color='black', fontsize=fontsize)

    elif add_y_median_labels or add_xy_median_labels:
        if continuous_x_col is None:
            raise ValueError(f"continuous_x_col must be given")
        df_by_x = df.groupby(x, sort=False)

        for x_idx, (key, g_data) in enumerate(df_by_x):
            y_val = g_data[y].median()
            x_val = g_data[continuous_x_col].median()
            if add_y_median_labels:
                ax.text(x_idx, y_val, f"{yvar_format.format(y_val)}\n\n",
                        ha='center', va='center',
                        color='black', fontsize=fontsize)
            else:
                ax.text(x_idx, y_val, f"med:\ny={yvar_format.format(y_val)}\nx={xvar_format.format(x_val)}",
                        ha='center', va='center',
                        color='black', fontsize=fontsize)

    # labels
    if xlabel is not None:
        if isinstance(xlabel, bool):
            if xlabel is True:
                xlabel = f"x = {x}"
            else:
                xlabel = None

    if ylabel is not None:
        if isinstance(ylabel, bool):
            if ylabel is True:
                ylabel = f"y = {y}"
            else:
                ylabel = None

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    # need to add line for legend with hues
    lines = None
    if labels is not None:
        lines = []
        for label, color in zip(labels, palette):
            lines.append((label, {'color': color}))

    if legend_loc is not None:
        put.set_legend(ax=ax,
                       legend_title=hue if add_hue_to_legend_title else None,
                       legend_loc=legend_loc,
                       fontsize=fontsize,
                       lines=lines,
                       **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)
    put.set_spines(ax=ax)

    return fig


def df_boxplot_by_index(df: Union[pd.Series, pd.DataFrame],
                        index_var_name: str = 'date',
                        ylabel: str = 'returns',
                        xlabel: str = None,
                        show_ylabel: bool = True,
                        title: str = None,
                        colors: Optional[List[str]] = None,
                        ax: plt.Subplot = None,
                        **kwargs
                        ) -> Optional[plt.Figure]:
    """
    plot data frame by aggregating by columns
    xvar is the index
    appropriate when y data is uniform conditional n regimes
    """
    # make sure it is given
    df.index.name = index_var_name
    box_data = dfm.melt_df_by_columns(df=df, x_index_var_name=index_var_name, y_var_name=ylabel)
    if colors is None:
        colors = put.compute_heatmap_colors(a=np.nanmean(df.to_numpy(), axis=1))

    fig = plot_box(df=box_data,
                   x=index_var_name,
                   y=ylabel,
                   continuous_x_col=ylabel,
                   ylabel=ylabel if show_ylabel else None,
                   original_index=df.index,
                   colors=colors,
                   xlabel=xlabel or False,
                   title=title,
                   ax=ax,
                   **kwargs)
    return fig


def df_boxplot_by_columns(df: Union[pd.Series, pd.DataFrame],
                          hue_var_name: str = 'instruments',
                          y_var_name: str = 'weights',
                          ylabel: str = 'weights',
                          show_ylabel: bool = True,
                          title: str = None,
                          colors: Optional[List[str]] = None,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> Optional[plt.Figure]:
    """
    plot boxplot using data by columns
    x-axis will be colum names
    """
    box_data = dfm.melt_df_by_columns(df=df, x_index_var_name=None, hue_var_name=hue_var_name, y_var_name=y_var_name)

    if colors is None:
        colors = put.compute_heatmap_colors(a=np.nanmean(df.to_numpy(), axis=0), **kwargs)

    fig = plot_box(df=box_data,
                   x=hue_var_name,
                   y=y_var_name,
                   continuous_x_col=y_var_name,
                   ylabel=ylabel if show_ylabel else None,
                   xlabel=hue_var_name,
                   original_index=None,
                   labels=df.columns.to_list(),
                   colors=colors,
                   title=title,
                   ax=ax,
                   **kwargs)

    return fig


def df_dict_boxplot_by_columns(dfs: Dict[str, Union[pd.Series, pd.DataFrame]],
                               hue_var_name: str = 'instruments',
                               y_var_name: str = 'weights',
                               ylabel: str = 'weights',
                               hue: str = 'Portfolio',
                               show_ylabel: bool = True,
                               title: str = None,
                               colors: Optional[List[str]] = None,
                               ax: plt.Subplot = None,
                               **kwargs
                               ) -> Optional[plt.Figure]:
    """
    dict keys are added as hue
    plot boxplot using data by columns
    x-axis will be colum names
    """
    box_datas = []
    for key, df in dfs.items():
        box_data = dfm.melt_df_by_columns(df=df, x_index_var_name=None, hue_var_name=hue_var_name, y_var_name=y_var_name)
        box_data[hue] = key
        box_datas.append(box_data)
    box_datas = pd.concat(box_datas, axis=0)
    if colors is None:
        colors = put.get_n_colors(n=len(dfs.keys()), **kwargs)

    fig = plot_box(df=box_datas,
                   x=hue_var_name,
                   y=y_var_name,
                   hue=hue,
                   labels=list(dfs.keys()),
                   continuous_x_col=y_var_name,
                   ylabel=ylabel if show_ylabel else None,
                   xlabel=hue_var_name,
                   original_index=None,
                   colors=colors,
                   title=title,
                   ax=ax,
                   **kwargs)

    return fig


def df_boxplot_by_hue_var(df: Union[pd.Series, pd.DataFrame],
                          hue_var_name: str = 'date',
                          x_index_var_name: str = None,
                          y_var_name: str = 'returns',
                          is_heatmap_colors: bool = True,
                          hue_order: List[str] = None,
                          title: str = None,
                          colors: List[str] = None,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> Optional[plt.Figure]:
    """
    plot time series of data frame by melting columns
    hue_var_name is hue index
    """
    box_data = dfm.melt_df_by_columns(df=df,
                                      x_index_var_name=x_index_var_name,
                                      y_var_name=y_var_name,
                                      hue_var_name=hue_var_name,
                                      hue_order=hue_order)
    if colors is None:
        if is_heatmap_colors:
            a = np.nanmedian(df.to_numpy(), axis=0)
            colors = put.compute_heatmap_colors(a=np.where(np.isfinite(a), a, 0.0))
        else:
            n = len(df.columns) if isinstance(df, pd.DataFrame) else 1
            colors = put.get_n_colors(n=n, **kwargs)

    fig = plot_box(df=box_data,
                   x=x_index_var_name,
                   y=y_var_name,
                   hue=hue_var_name,
                   colors=colors,
                   title=title,
                   ax=ax,
                   **kwargs)

    return fig


def df_boxplot_by_classification_var(df: pd.DataFrame,
                                     x: str,  # classification variables
                                     y: str,
                                     num_buckets: int = 6,
                                     x_hue_name: str = 'x_regimes',
                                     hue: Optional[str] = None,
                                     hue_order: List[str] = None,
                                     title: str = None,
                                     is_add_xlabel: bool = True,
                                     xvar_format: str = '{:.2f}',
                                     yvar_format: str = '{:.2f}',
                                     showfliers: bool = False,
                                     showmeans: bool = False,
                                     meanline: bool = False,
                                     medianline: bool = True,
                                     add_xy_mean_labels: bool = False,
                                     add_xy_med_labels: bool = False,
                                     is_value_labels: bool = True,
                                     colors: Union[List[str], List[Tuple[float, float, float]]] = None,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
    """
    use x as classification var and plod box plot relative to quatiles of x
    """
    df = df.dropna()
    scatter_data, _ = dfc.add_quantile_classification(df=df, x_column=x,
                                                      hue_name=x_hue_name,
                                                      num_buckets=num_buckets,
                                                      bins=None,
                                                      xvar_format=xvar_format,
                                                      is_value_labels=is_value_labels)
    if hue_order is not None:
        labels = hue_order
    else:
        labels = None

    plot_box(df=scatter_data,
             x=x_hue_name,
             y=y,
             hue=hue,
             continuous_x_col=x,
             xlabel=x_hue_name if is_add_xlabel else None,
             hue_order=hue_order,
             title=title,
             xvar_format=xvar_format,
             yvar_format=yvar_format,
             showfliers=showfliers,
             showmeans=showmeans,
             meanline=meanline,
             medianline=medianline,
             add_xy_mean_labels=add_xy_mean_labels,
             add_xy_median_labels=add_xy_med_labels,
             labels=labels,
             colors=colors,
             ax=ax,
             **kwargs)


def df_dict_boxplot_by_classification_var(data_dict: Dict[Tuple[str, str], pd.DataFrame],
                                          num_buckets: int = 6,
                                          x_hue_name: str = 'x_regimes',
                                          var_hue_name: str = 'vars',
                                          y_var_name: str = 'y_var',
                                          title: str = None,
                                          is_add_xlabel: bool = True,
                                          xvar_format: str = '{:.2f}',
                                          yvar_format: str = '{:.2f}',
                                          showfliers: bool = False,
                                          showmeans: bool = False,
                                          meanline: bool = False,
                                          medianline: bool = True,
                                          add_xy_mean_labels: bool = False,
                                          add_xy_med_labels: bool = False,
                                          colors: Union[List[str], List[Tuple[float, float, float]]] = None,
                                          is_value_labels: bool = False,
                                          is_add_last_value: bool = False,
                                          ax: plt.Subplot = None,
                                          **kwargs
                                          ) -> None:
    """
    get dict of dfs and add regime vars, key = (x_column, y_column)
    use x as classification var and plod box plot relative to quatiles of x
    is_value_labels will be same across different df and x
    """
    scatter_datas = None
    hue_order = []
    for key, df in data_dict.items():
        scatter_data, _ = dfc.add_quantile_classification(df=df, x_column=key[0],
                                                          hue_name=x_hue_name,
                                                          num_buckets=num_buckets,
                                                          bins=None,
                                                          xvar_format=xvar_format,
                                                          is_value_labels=is_value_labels,
                                                          **kwargs)
        scatter_data[var_hue_name] = key[1]  # y is id
        scatter_data = scatter_data.rename({key[1]: y_var_name}, axis=1)
        if scatter_datas is None:
            scatter_datas = scatter_data
        else:
            scatter_datas = pd.concat([scatter_datas, scatter_data], axis=0)
        hue_order.append(key[1])
    plot_box(df=scatter_datas,
             x=x_hue_name,
             y=y_var_name,
             hue=var_hue_name,
             xlabel=x_hue_name if is_add_xlabel else None,
             hue_order=hue_order,
             title=title,
             xvar_format=xvar_format,
             yvar_format=yvar_format,
             showfliers=showfliers,
             showmeans=showmeans,
             meanline=meanline,
             medianline=medianline,
             add_xy_mean_labels=add_xy_mean_labels,
             add_xy_median_labels=add_xy_med_labels,
             labels=hue_order,
             colors=colors,
             ax=ax,
             **kwargs)
    if is_add_last_value:
        scatter_datas = scatter_datas.sort_index().ffill()
        last = scatter_datas.iloc[-1, :]
        for idx, key in enumerate(data_dict.keys()):
            ax.annotate('Last',
                        xy=(last[x_hue_name], last[key[1]]), xytext=(1, 1),
                        textcoords='offset points', ha='left', va='bottom',
                        fontsize=12)
            ax.scatter(x=last[x_hue_name], y=last[key[1]], s=20)


class LocalTests(Enum):
    RETURNS_BOXPLOT = 1
    DF_BOXPLOT = 2
    DF_BOXPLOT_INDEX = 3
    DF_WEIGHTS = 4
    DF_DICT = 5


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data()
    returns = prices.asfreq('QE', method='ffill').pct_change()

    if local_test == LocalTests.RETURNS_BOXPLOT:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)

        index_name = 'date'
        value_name = 'returns'

        box_data = dfm.melt_df_by_columns(df=returns, x_index_var_name=index_name, y_var_name=value_name)
        print(box_data)
        colors = put.compute_heatmap_colors(a=np.nanmean(returns.to_numpy(), axis=1))

        plot_box(df=box_data,
                 x=index_name,
                 y=value_name,
                 original_index=returns.index,
                 colors=colors,
                 xlabel=False,
                 ax=ax,
                 **global_kwargs)

    elif local_test == LocalTests.DF_BOXPLOT:
        # returns by the quantiles of the first variable
        var = returns.columns[0]
        df_boxplot_by_classification_var(df=returns[var].to_frame(), x=var, y=var)

    elif local_test == LocalTests.DF_BOXPLOT_INDEX:
        df_boxplot_by_index(df=returns)

    elif local_test == LocalTests.DF_WEIGHTS:
        df_boxplot_by_columns(df=prices,
                              hue_var_name='instruments',
                              y_var_name='weights',
                              ylabel='weights',
                              legend_loc=None,
                              showmedians=True,
                              add_y_median_labels=True)

    elif local_test == LocalTests.DF_DICT:
        dfs = {'alts': prices, 'bal': 0.5*prices}
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(8, 6))
            df_dict_boxplot_by_columns(dfs=dfs,
                                       hue_var_name='instruments',
                                       y_var_name='weights',
                                       ylabel='weights',
                                       legend_loc='upper center',
                                       showmedians=True,
                                       add_y_median_labels=True,
                                       ncols=2,
                                       ax=ax)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.DF_DICT)



"""
heatmap plots
"""
# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from typing import List, Optional, Union
from enum import Enum

# qis
import qis.plots.utils as put


def plot_heatmap(df: pd.DataFrame,
                 transpose: bool = False,
                 inverse: bool = False,
                 date_format: Optional[str] = '%Y',
                 cmap: Union[str, ListedColormap] = 'RdYlGn',
                 var_format: Optional[str] = '{:.1%}',
                 alpha: float = 1.0,
                 fontsize: int = 10,
                 title: Optional[str] = None,
                 top_x_label: bool = True,
                 square: bool = False,
                 vline_columns: List[int] = None,
                 hline_rows: List[int] = None,
                 vmin: float = None,
                 vmax: float = None,
                 labelpad: int = 50,
                 ylabel: str = '',
                 annot: Union[bool, np.ndarray] = True,
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> Optional[plt.Figure]:

    if ax is None:
        fig, ax = plt.subplots()
    else:  # add table to existing axis
        fig = None

    df = df.copy()

    if date_format is not None:  # index may include 'Total'
        df.index = [date.strftime(date_format) if isinstance(date, pd.Timestamp) else date for date in df.index]

    if transpose:
        df = df.T
        inverse = False

    if inverse:
        df = df.reindex(index=df.index[::-1])

    if var_format is not None:
        var_format = var_format.replace('{:', '').replace('}', '')  # no {}
    else:
        var_format = ""

    sns.heatmap(data=df,
                center=0,
                annot=annot,
                fmt=var_format,
                cmap=cmap,
                alpha=alpha,
                cbar_kws={'size': fontsize},
                cbar=False,
                annot_kws={'size': fontsize},
                xticklabels=True,  # important for full display of labels
                yticklabels=True,  # important for full display of labels
                square=square,
                vmin=vmin,
                vmax=vmax,
                ax=ax)  # ,"ha": 'right' #cbar_kws={'format': '%0.2f%%'}

    if top_x_label:
        ax.xaxis.tick_top()

    if not transpose:
        pass
        # bottom, top = ax.get_ylim()
        # ax.set_ylim(bottom + 0.5, top - 0.5)
    else:
        ax.xaxis.labelpad = labelpad

    put.set_ax_tick_params(ax=ax, fontsize=fontsize, labelbottom=not top_x_label, labeltop=top_x_label, **kwargs)
    put.set_ax_tick_labels(ax=ax, fontsize=fontsize, **kwargs)

    if vline_columns is not None:
        for vline_column in vline_columns:
            ax.vlines([vline_column], *ax.get_ylim(), lw=1)

    if hline_rows is not None:
        for hline_row in hline_rows:
            ax.hlines([hline_row], *ax.get_xlim(), lw=1)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    ax.set_ylabel(ylabel)
    ax.set_xlabel('')

    return fig


class LocalTests(Enum):
    HEATMAP = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if local_test == LocalTests.HEATMAP:
        corrs = prices.pct_change().corr()
        plot_heatmap(corrs, inverse=False, x_rotation=90)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.HEATMAP)



"""
Plot regime-conditional regression analysis with scatter plots and fitted lines.

Visualizes asset returns against benchmark returns with regime-specific betas,
showing how asset sensitivities vary across different market conditions.
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum
from matplotlib.ticker import FuncFormatter
from typing import Optional

# qis
import qis.plots.utils as put
import qis.perfstats.cond_regression as cre
from qis.perfstats.regime_classifier import (RegimeClassifier,
                                             BenchmarkReturnsQuantilesRegime)


def plot_scatter_regression(prices: pd.DataFrame,
                            regime_benchmark: str,
                            regime_classifier: Optional[RegimeClassifier] = None,
                            drop_benchmark: bool = True,
                            is_add_alpha: bool = True,
                            x_var_format: str = '{:.0%}',
                            y_var_format: str = '{:.0%}',
                            beta_format: str = '{:.2f}',
                            xlabel: Optional[str] = None,
                            ylabel: Optional[str] = None,
                            title: Optional[str] = None,
                            display_estimated_betas: bool = True,
                            is_print_summary: bool = False,
                            add_last_date: bool = False,
                            framealpha: float = 0.9,
                            ax: Optional[plt.Subplot] = None,
                            **kwargs
                            ) -> Optional[plt.Figure]:
    """
    Plot scatter regression with regime-conditional betas.

    Creates a scatter plot showing asset returns vs benchmark returns, with fitted
    regression lines that have different slopes (betas) for each market regime.
    Background colors indicate regime boundaries.

    Args:
        prices: DataFrame of asset prices with benchmark included
        regime_benchmark: Column name of benchmark asset for regime classification
        regime_classifier: Regime classifier instance (default: BenchmarkReturnsQuantilesRegime with default params)
        drop_benchmark: If True, exclude benchmark from regression outputs
        x_var_format: Format string for x-axis values (benchmark returns)
        y_var_format: Format string for y-axis values (asset returns)
        beta_format: Format string for displaying beta values in legend
        xlabel: Custom x-axis label (default: "x={benchmark}")
        ylabel: Custom y-axis label (default: "Assets" or specific asset name)
        title: Plot title (default: None)
        display_estimated_betas: If True, show detailed regime betas in legend
        is_print_summary: If True, print OLS regression summaries
        framealpha: Legend frame transparency (0=transparent, 1=opaque)
        add_last_date: If True, add annotation for most recent data point
        ax: Matplotlib axis to plot on (creates new figure if None)
        **kwargs: Additional arguments passed to plotting utilities

    Returns:
        Figure object if new figure created, None if plotting on existing axis

    Examples:
        # Plot with quantile regimes (default)
        fig = plot_scatter_regression(
            prices=prices,
            regime_benchmark='SPY',
            regime_classifier=BenchmarkReturnsQuantilesRegime(freq='QE')
            ),
            is_asset_detailed=True
        )

        # Plot with positive/negative regimes
        fig = plot_scatter_regression(
            prices=prices,
            regime_benchmark='SPY',
            regime_classifier=BenchmarkReturnsPositiveNegativeRegime(freq='ME')
            ),
            is_asset_detailed=True
        )

        # Plot with default settings
        fig = plot_scatter_regression(
            prices=prices[['SPY', 'TLT']],
            regime_benchmark='SPY',
            add_last_date=True
        )
    """
    # Create new figure if axis not provided
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    # Initialize regime classifier with default if not provided
    if regime_classifier is None:
        regime_classifier = BenchmarkReturnsQuantilesRegime()

    # Estimate regime-conditional regression parameters
    regmodel_out_dict = {}
    estimated_params = cre.estimate_cond_regression(
        prices=prices,
        benchmark=regime_benchmark,
        drop_benchmark=drop_benchmark,
        regime_classifier=regime_classifier,  # Access params from classifier
        is_add_alpha=is_add_alpha,
        is_print_summary=is_print_summary,
        regmodel_out_dict=regmodel_out_dict
    )

    # Get unique markers and colors for each asset
    n_assets = len(estimated_params.index)
    markers = put.get_n_markers(n=n_assets)
    colors = put.get_n_colors(n=n_assets)

    # Build legend entries
    lines = []

    # Plot scatter and regression lines for each asset
    for (asset, pandas_out), marker, color_ in zip(regmodel_out_dict.items(), markers, colors):
        # Sort by benchmark for proper line plotting
        pandas_out = pandas_out.sort_values(regime_benchmark)

        # Determine scatter point coloring strategy
        if display_estimated_betas and len(prices.columns) == 1:
            # Single asset: use regime colors for scatter points
            scatter_hue = pandas_out[regime_classifier.REGIME_COLUMN]
            scatter_palette = list(regime_classifier.get_regime_ids_colors().values())
            scatter_hue_order = list(regime_classifier.get_regime_ids_colors().keys())
            scatter_color = None
        else:
            # Multiple assets: use asset-specific colors
            scatter_hue = None
            scatter_palette = None
            scatter_hue_order = None
            scatter_color = color_

        # Plot scatter points for actual returns
        sns.scatterplot(
            x=regime_benchmark,
            y=asset,
            hue=scatter_hue,
            data=pandas_out,
            palette=scatter_palette,
            hue_order=scatter_hue_order,
            color=scatter_color,
            marker=marker,
            legend=None,
            ax=ax
        )

        # Configure legend and line styling based on detail level
        # Configure legend and line styling based on detail level
        if display_estimated_betas:
            # Detailed view: show regime-specific betas
            legend = True
            line_marker = None
            line_hue = pandas_out[regime_classifier.REGIME_COLUMN]
            line_hue_order = list(regime_classifier.get_regime_ids_colors().keys())

            if len(prices.columns) == 1:
                # Single asset: separate legend entry per regime
                palette = list(regime_classifier.get_regime_ids_colors().values())
                for regime, color in regime_classifier.get_regime_ids_colors().items():
                    lines.append((
                        f"{regime} beta={beta_format.format(estimated_params.loc[asset, regime])}",
                        {'color': color, 'linestyle': '-', 'marker': line_marker}
                    ))
            else:
                # Multiple assets: combined legend entry with all betas
                palette = [color_] * len(regime_classifier.get_regime_ids_colors().values())
                label = f"{asset}: "
                beta_parts = [
                    f"{regime} beta={beta_format.format(estimated_params.loc[asset, regime])}"
                    for regime in regime_classifier.get_regime_ids_colors().keys()
                ]
                label += ", ".join(beta_parts)
                lines.append((label, {'color': color_, 'linestyle': '-', 'marker': marker}))
        else:
            # Simple view: one legend entry per asset, single color line
            palette = None
            legend = None
            line_marker = marker
            line_hue = None
            line_hue_order = None
            lines.append((asset, {'color': color_, 'linestyle': '-', 'marker': line_marker}))

        # Plot fitted regression lines
        sns.lineplot(
            x=regime_benchmark,
            y=cre.ConditionalRegressionColumns.PREDICTION.value,
            hue=line_hue,
            data=pandas_out,
            hue_order=line_hue_order,
            palette=palette,
            color=color_ if not display_estimated_betas else None,  # Use direct color when no hue
            marker=line_marker,
            legend=legend,
            ax=ax
        )

    # Add colored background regions indicating market regimes
    regime_colors = list(regime_classifier.get_regime_ids_colors().values())
    regime_ids = list(regime_classifier.get_regime_ids_colors().keys())

    # Extract regime boundaries from first asset (all share same benchmark)
    for asset, pandas_out in regmodel_out_dict.items():
        # Find min/max benchmark returns for each regime
        regime_boundaries = {}
        for regime in regime_ids:
            regime_data = pandas_out[pandas_out[regime_classifier.REGIME_COLUMN] == regime]
            if not regime_data.empty:
                regime_boundaries[regime] = (
                    regime_data[regime_benchmark].min(),
                    regime_data[regime_benchmark].max()
                )

        # Save current axis limits before adding background
        axis_x_min, axis_x_max = ax.get_xlim()
        axis_y_min, axis_y_max = ax.get_ylim()

        # Sort regimes by x position for proper ordering
        sorted_regimes = sorted(regime_boundaries.items(), key=lambda x: x[1][0])

        # Draw vertical spans for each regime
        last_x_min = axis_x_min
        for idx, (regime, (x_min, x_max)) in enumerate(sorted_regimes):
            regime_idx = regime_ids.index(regime)

            # Extend first regime to axis minimum, last regime to axis maximum
            span_x_min = last_x_min # axis_x_min if idx == 0 else x_min
            span_x_max = axis_x_max if idx == len(sorted_regimes) - 1 else x_max
            last_x_min = span_x_max
            ax.axvspan(span_x_min, span_x_max, alpha=0.15, color=regime_colors[regime_idx], zorder=0)

        # Restore original axis limits (axvspan can modify them)
        ax.set_xlim(axis_x_min, axis_x_max)
        ax.set_ylim(axis_y_min, axis_y_max)

        break  # Only need to process once since all assets share same benchmark

    # Add annotation for most recent data point if requested
    if add_last_date:
        label_x_y = {}
        for asset, df in regmodel_out_dict.items():
            x = df[regime_benchmark].iloc[-1]
            y = df[asset].iloc[-1]
            label = (
                f"Last {df.index[-1].strftime('%d-%b-%Y')}: "
                f"x={x_var_format.format(x)}, y={x_var_format.format(y)}"
            )
            label_x_y[label] = (x, y)

        # Determine colors for annotation points
        if len(prices.columns) == 2:
            annotation_colors = [df[cre.COLOR_COLUMN].iloc[-1]]
        else:
            annotation_colors = colors

        put.add_scatter_points(ax=ax, label_x_y=label_x_y, colors=annotation_colors, **kwargs)

    # Configure plot aesthetics
    put.set_legend(ax=ax, lines=lines, framealpha=framealpha, **kwargs)
    ax.xaxis.set_major_formatter(FuncFormatter(lambda x, _: x_var_format.format(x)))
    ax.yaxis.set_major_formatter(FuncFormatter(lambda y, _: y_var_format.format(y)))

    # Set axis labels with defaults
    if xlabel is None:
        xlabel = f"x={regime_benchmark}"
    if ylabel is None:
        ylabel = f"y = {asset}" if len(prices.columns) == 2 else 'Assets'
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, **kwargs)

    # Set title if provided
    if title is not None:
        put.set_title(ax=ax, title=title, **kwargs)

    put.set_spines(ax=ax, **kwargs)

    return fig



"""
file utils to work with sql_engine
requires SQLAlchemy>=2.0.0
"""


import pandas as pd
import yaml
from sqlalchemy import create_engine, Engine
from pathlib import Path
from qis.file_utils import timer, INDEX_COLUMN
from enum import Enum
from typing import Dict, Union, NamedTuple, Optional, Literal, List


def get_engine(path: str = 'AWS_POSTGRES') -> Engine:
    full_file_path = Path(__file__).parent.joinpath('settings.yaml')
    with open(full_file_path) as settings:
        settings_data = yaml.load(settings, Loader=yaml.Loader)
    path = settings_data[path]
    engine = create_engine(path)
    return engine


@timer
def save_df_dict_to_sql(engine: Engine,
                        table_name: str,
                        dfs: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                        schema: Optional[str] = None,
                        index_col: Optional[str] = INDEX_COLUMN,
                        if_exists: Literal["fail", "replace", "append"] | Literal["truncate-append"] = "fail",
                        ) -> None:
    """
    save pandas dict to sql engine
    """
    for key, df in dfs.items():
        if df is not None and isinstance(df, pd.DataFrame):
            if index_col is not None:
                df = df.reset_index(names=index_col)
            if if_exists == "truncate-append":
                schema_str = f"{schema}." if schema else ""
                with engine.connect() as con:
                    statement = f"TRUNCATE TABLE {schema_str}{table_name}_{key}"
                    con.execute(statement)
                    con.commit()
                df.to_sql(
                    f"{table_name}_{key}",
                    engine,
                    schema=schema,
                    if_exists='append',
                    method='multi',
                    chunksize=1000
                )
            else:
                df.to_sql(f"{table_name}_{key}", engine, schema=schema, if_exists=if_exists)


@timer
def load_df_dict_from_sql(engine: Engine,
                          table_name: str,
                          dataset_keys: List[Union[str, Enum, NamedTuple]],
                          schema: Optional[str] = None,
                          index_col: Optional[str] = INDEX_COLUMN,
                          columns: Optional[List[str]] = None,
                          drop_sql_index: bool = True
                          ) -> Dict[str, pd.DataFrame]:
    """
    pandas dict from csv files
    """
    pandas_dict = {}
    for key in dataset_keys:
        # df will have index set by index_col with added column 'index' from sql
        df = pd.read_sql_table(table_name=f"{table_name}_{key}", con=engine, schema=schema,
                               index_col=index_col,
                               columns=columns)
        if drop_sql_index:
            df = df.drop('index', axis=1)
            #  df[index_col] = pd.to_datetime(df[index_col])
            #  df = df.set_index(index_col, drop=True)
        pandas_dict[key] = df
    return pandas_dict



"""
plot df as table
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.table import Table as Table
from enum import Enum
from typing import List, Tuple, Optional, Literal, Union

# qis
import qis.plots.utils as put
import qis.utils.df_str as dfs

ROW_HIGHT = 0.625  # cm
COLUMN_WIDTH = 2.0  # cm ?
FIRST_COLUMN_WIDTH = 3.0


def plot_df_table(df: Union[pd.DataFrame, pd.Series],
                  add_index_as_column: bool = True,
                  column_width: float = COLUMN_WIDTH,
                  row_height: float = ROW_HIGHT,
                  first_column_width: Optional[float] = FIRST_COLUMN_WIDTH,
                  first_row_height: float = None,
                  col_widths: List[float] = None,  # can pass as cols
                  rotation_for_columns_headers: int = None,
                  rotation_for_text: int = None,
                  transpose: bool = False,
                  index_column_name: str = ' ',
                  fontsize: int = 10,
                  header_color: str = '#40466e',
                  header_text_color: str = 'w',
                  row_colors: List[str] = ('#f1f1f2', 'w'),
                  edge_color: str = 'lightgray',
                  bbox: Tuple[float] = (0, 0, 1, 1),
                  header_column_id: int = 0,
                  header_row_id: int = 0,
                  left_aligned_first_col: bool = False,
                  var_format: str = None,  # '{:.2f}' to convert numerical data to str
                  title: str = None,
                  heatmap_columns: List[int] = None,
                  heatmap_rows: List[int] = None,
                  heatmap_rows_columns: Tuple[Tuple[int, int], Tuple[int, int]] = None,  # row[0]: row[1], column[0]:column[1]
                  cmap: str = 'RdYlGn',
                  special_rows_colors: List[Tuple[int, str]] = None,
                  special_columns_colors: List[Tuple[int, str]] = None,
                  data_colors: List[Tuple[float, float, float]] = None,
                  diagonal_color: str = None,
                  rows_edge_lines: List[int] = None,
                  rows_edge_color: str = 'blue',
                  columns_edge_lines: List[Tuple[int, str]] = None,
                  bold_font: bool = False,
                  linewidth: float = 0.5,  # table borders
                  alpha: float = 1.0,
                  emply_column_names: bool = False,
                  ax: plt.Subplot = None,
                  **kwargs
                  ) -> Optional[plt.Figure]:
    """
    plot dataframe as maotplotlib table
    """
    df = df.copy()  # data object will be changed
    if isinstance(df, pd.Series):
        df = df.to_frame()
    if transpose:
        t_data = df.T
        if add_index_as_column:
            t_data.columns = t_data.iloc[0, :]  # rename columns by original index
            df = t_data.drop(labels=t_data.index[0], axis=0)
        else:
            t_data.columns = df.index  # rename columns using original index
            df = t_data

    if add_index_as_column:
        df.insert(0, column=index_column_name, value=df.index)

    if col_widths is None:
        col_widths = [column_width for _ in df]
    else:
        first_column_width = col_widths[0]

    # allocate size
    size = (np.array(df.shape[::-1]) + np.array([0, 1])) * np.array([col_widths[0], row_height])

    if first_column_width is None:
        first_column_width = col_widths[0]
    #else:
    #    col_widths[0] = first_column_width # after this change

    if ax is None:  # create new axis
        fig, ax = plt.subplots(figsize=size)
        ax.axis('off')
    else:  # add table to existing axis
        fig = None
        ax.axis('off')

    if var_format is not None:
        df = dfs.df_to_str(df=df, var_format=var_format)

    if emply_column_names is False:
        col_labels = df.columns.to_list()
    else:
        col_labels = [df.columns[0]] + [''] * (len(df.columns) - 1)

    mpl_table = ax.table(cellText=df.to_numpy(),
                         bbox=bbox,
                         colLabels=col_labels,
                         colLoc='center')

    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(fontsize)

    weight = 'bold' if bold_font else 'normal'
    for k, cell in mpl_table._cells.items():
        cell.set_linewidth(0.5)

        if k[0] == header_row_id or k[1] < header_column_id:
            cell.set_text_props(weight=weight, color=header_text_color)
            cell.set_facecolor(header_color)
            if first_row_height is not None:
                cell.set_height(first_row_height)

            if rotation_for_columns_headers is not None:
                cell.get_text().set_rotation(rotation_for_columns_headers)  # set rotation
                # cell.set_height(first_column_width / 15)  # aling widths

        elif k[0] > header_row_id:
            cell.set_facecolor(row_colors[k[0] % len(row_colors) ])

        if k[1] == 0:
            width = first_column_width
        else:
            width = col_widths[k[1]]
        cell.set_width(width)
        if rotation_for_text is not None:
            cell.get_text().set_rotation(rotation_for_text)  # set rotation

    if heatmap_columns is not None:
        for heatmap_column in heatmap_columns:
            column_data = df[df.columns[heatmap_column]]
            if isinstance(column_data, pd.DataFrame):
                raise ValueError(f"dublicated columns: {column_data.columns}")
            x_array = dfs.series_to_numeric(ds=column_data)
            colors = put.compute_heatmap_colors(a=x_array, cmap=cmap)
            for k, cell in mpl_table._cells.items():
                if k[0] > 0 and k[1] == heatmap_column:  # skip first row
                    cell.set_facecolor(colors[k[0]-1])

    if heatmap_rows is not None:
        for heatmap_row in heatmap_rows:
            row_data = df.iloc[heatmap_row, 1:]  # exclude first row
            x_array = dfs.series_to_numeric(ds=row_data)
            colors = put.compute_heatmap_colors(a=x_array, cmap=cmap)
            for k, cell in mpl_table._cells.items():
                if k[1] > 0 and k[0] == heatmap_row+1:  # heatmap_row is not counting first headers
                    cell.set_facecolor(colors[k[1]-1])

    if heatmap_rows_columns is not None:   # row[0]: row[1], column[0]:column[1]
        # set colors on all data
        row_start, row_end = heatmap_rows_columns[0][0], heatmap_rows_columns[0][1]
        col_start, col_end = heatmap_rows_columns[1][0], heatmap_rows_columns[1][1]

        if add_index_as_column:
            col_start = col_start + 1
            col_end = col_end + 1
        else:
            col_start = col_start - 1

        data_extract = df.iloc[row_start: row_end, col_start: col_end]
        a = dfs.df_to_numeric(df=data_extract)
        colors = put.compute_heatmap_colors(a=a, cmap=cmap)
        for row_idx in range(len(data_extract.index)):
            col_idx = 0
            for k, cell in mpl_table._cells.items():
                if k[0] == row_start + row_idx + 1 and k[1] >= col_start and k[1] < col_end:
                    cell.set_facecolor(colors[row_idx][col_idx])
                    col_idx += 1

    if special_rows_colors is not None:
        for special_rows_color in special_rows_colors:
            set_cells_facecolor(mpl_table,
                                row=special_rows_color[0],
                                color=special_rows_color[1],
                                alpha=0.5*alpha,
                                bold_font=bold_font)

    if special_columns_colors is not None:
        for special_columns_color in special_columns_colors:
            set_cells_facecolor(mpl_table,
                                column=special_columns_color[0],
                                color=special_columns_color[1],
                                alpha=0.5*alpha,
                                bold_font=bold_font,
                                header_row_id=header_row_id)

    if diagonal_color is not None:
        set_diag_cells_facecolor(mpl_table, color=diagonal_color)

    if data_colors is not None:
        set_data_colors(mpl_table, header_column_id=0 if add_index_as_column else -1,
                        data_colors=data_colors)

    set_row_edge_color(mpl_table, row=None, color=edge_color)

    if left_aligned_first_col:
        set_align_for_column(mpl_table, col=0, align='left')
    else:
        set_align_for_column(mpl_table, col=0, align='right')

    if rows_edge_lines is not None or columns_edge_lines is not None:
        ax.axis(xmin=0, xmax=df.shape[1], ymin=df.shape[0], ymax=-1)  # need to reset axis to match rows position

    if rows_edge_lines is not None:
        for rows_edge_line in rows_edge_lines:
            ax.axhline(y=rows_edge_line, color=rows_edge_color, alpha=0.5*alpha, lw=0.75)

    if columns_edge_lines is not None:
        for columns_edge_line in columns_edge_lines:
            ax.axvline(x=columns_edge_line[0], color=columns_edge_line[1], alpha=0.5*alpha, lw=0.75)
            # set_column_edge_color(mpl_table, column=columns_edge_line[0], color=columns_edge_line[1])

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    return fig


def plot_df_table_with_ci(df: pd.DataFrame,
                          df_ci: pd.DataFrame,
                          var_format: str = '{:.2f}',
                          ax: plt.Subplot = None,
                          is_add_heatmap: bool = True,
                          axis: Literal[None, 0, 1] = 1,  # heatmap by column
                          **kwargs
                          ) -> plt.Figure:
    """
    table with ci
    """
    table_str = dfs.df_with_ci_to_str(df=df,
                                      df_ci=df_ci,
                                      var_format=var_format)
    if is_add_heatmap:
        data_colors = put.compute_heatmap_colors(a=df.to_numpy(), axis=axis)
    else:
        data_colors = None

    fig = plot_df_table(df=table_str,
                        data_colors=data_colors,
                        ax=ax,
                        **kwargs)
    return fig


def set_row_edge_color(table: Table,
                       row: int = None,
                       color: str = 'slategray'
                       ) -> None:
    for k, cell in table._cells.items():
        if row is None:
            cell.set_edgecolor(color)
        else:
            if k[0] == row:
                cell.set_edgecolor(color)


def set_column_edge_color(table: Table,
                          column: int = None,
                          color: str = 'slategray'
                          ) -> None:
    for k, cell in table._cells.items():
        if column is None:
            cell.set_edgecolor(color)
        else:
            if k[1] == column:
                cell.set_edgecolor(color)
                cell.visible_edges = "L"


def set_cells_facecolor(table: Table,
                        header_row_id: int = 0,
                        row: int = None,
                        column: int = None,
                        color: str = 'slategray',
                        alpha: float = 1.0,
                        bold_font: bool = False
                        ) -> None:

    for k, cell in table._cells.items():
        if row is None and column is None:
            cell.set_facecolor(color)
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")

        elif column is None:  # set solor to row
            if k[0] == row:
                cell.set_facecolor(color)
                if bold_font:
                    txt = cell.get_text()
                    txt.set_fontweight("bold")

        elif row is None:  # set solor to column
            if k[1] == column and k[0] > header_row_id:
                cell.set_facecolor(color)
                cell.set_alpha(alpha)
                if bold_font:
                    txt = cell.get_text()
                    txt.set_fontweight("bold")


def set_diag_cells_facecolor(table: Table,
                             color: str = 'slategray',
                             bold_font: bool = False
                             ) -> None:
    for k, cell in table._cells.items():
        if k[0] == k[1] and k[0] > 0:
            cell.set_facecolor(color)
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")


def set_data_colors(table: Table,
                    data_colors: np.ndarray,
                    header_row_id: int = 0,
                    header_column_id = 0,
                    bold_font: bool = False
                    ) -> None:
    for k, cell in table._cells.items():
        if k[1] > header_column_id and k[0] > header_row_id:
            cell.set_facecolor(data_colors[k[0]-1][k[1]-1])
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")


def set_align_for_column(table: Table,
                         col: int,
                         align: str = 'left'
                         ) -> None:
    cells = [key for key in table._cells if key[1] == col]
    for cell in cells:
        table.properties()["celld"][cell]._loc = align


class LocalTests(Enum):
    TABLE = 0


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.TABLE:

        cars = {'Brand': ['Honda Civic', 'Toyota Corolla', 'Ford Focus', 'Audi A4'],
                'Price': [220.0, 250.0, 270.0, 35.0],
                'Engine': [175.0, 300.0, 100.0, 500.0],
                'Speed': [200.0, 150.0, 200.0, 175.0]}

        data = pd.DataFrame.from_dict(cars)
        data = data.set_index('Brand', drop=False)
        print(data)
        data['Price'] = data['Price']
        data['Engine'] = data['Engine']
        plot_df_table(df=data, heatmap_columns=[2], bold_font=False)
        plot_df_table(df=data, heatmap_rows_columns=((0, len(data.index)), (3, 4)))

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.TABLE)



"""
Regime-conditional regression analysis for portfolio factor exposure estimation.

This module implements regime-conditional regression where asset returns are regressed
on regime-specific benchmark returns. The key innovation is partitioning the sample
by market regimes (e.g., bear/normal/bull) and estimating separate factor loadings
(betas) for each regime.

Regression specification (without alpha):
    R_i,t = β_i,bear * R_b,t * I_bear + β_i,normal * R_b,t * I_normal + β_i,bull * R_b,t * I_bull + ε_i,t

where:
    - R_i,t: asset i returns at time t
    - R_b,t: benchmark returns at time t
    - I_regime: indicator function (1 if regime active, 0 otherwise)
    - β_i,regime: asset i's beta exposure in each regime
    - ε_i,t: residual (idiosyncratic) returns

Alpha term consideration (is_add_alpha parameter):
    - is_add_alpha=False (RECOMMENDED): Regime-specific betas capture all systematic variation.
      No constant term needed as regime dummies span the full sample space.
    - is_add_alpha=True: Adds global constant α_i, but creates interpretation issues:
      * What does constant mean when you have regime-specific exposures?
      * May cause multicollinearity as regime dummies already partition sample
      * Alpha becomes unidentified or captures numerical artifacts

Standard practice in regime-switching models excludes the global alpha term, letting
the regime-conditional betas fully characterize the asset's behavior across market states.

Alternative: For regime-specific alphas, extend model to:
    R_i,t = α_i,bear * I_bear + β_i,bear * R_b,t * I_bear + ... + ε_i,t
"""

# built-in
import numpy as np
import pandas as pd
import statsmodels.api as sm
from enum import Enum
from typing import Dict, Optional

# qis
from qis.perfstats.regime_classifier import (BenchmarkReturnsQuantilesRegime,
                                             RegimeClassifier)

class ConditionalRegressionColumns(str, Enum):
    # Column name constants
    RESID_VAR_COLUMN = 'An var resid'
    ALPHA_COLUMN = 'Alpha'
    COLOR_COLUMN = 'Color'
    PREDICTION = 'Prediction'
    VAR = 'Var'
    R2 = 'R2'


def estimate_cond_regression(prices: pd.DataFrame,
                             benchmark: str,
                             regime_classifier: Optional[RegimeClassifier] = None,
                             is_print_summary: bool = False,
                             drop_benchmark: bool = True,
                             min_period: int = 0,
                             is_add_alpha: bool = False,
                             regmodel_out_dict: Optional[Dict[str, pd.DataFrame]] = None
                             ) -> pd.DataFrame:
    """
    Estimate regime-conditional regression parameters (betas) for multiple assets.

    Fits separate betas for each market regime (e.g., bear, normal, bull) by regressing
    asset returns on regime-specific benchmark returns.

    Args:
        prices: DataFrame of asset prices with benchmark as first column
        benchmark: Name of benchmark column used for regime classification
        regime_classifier: Specifications for regime classification (quantiles, frequency)
        is_print_summary: If True, print OLS regression summary for each asset
        drop_benchmark: If True, exclude benchmark from regression outputs
        min_period: Minimum number of valid returns required to estimate parameters
        is_add_alpha: If True, include intercept term in regression
        regmodel_out_dict: Optional dict to store detailed regression outputs per asset

    Returns:
        DataFrame with regime betas (and alpha if requested) as columns, assets as rows

    Raises:
        ValueError: If benchmark has all NaN returns (invalid benchmark specification)
    """
    # Initialize regime classifier with default params if not provided

    if regime_classifier is None:
        regime_classifier = BenchmarkReturnsQuantilesRegime()

    # Classify returns into market regimes based on benchmark quantiles
    sampled_returns_with_regime_id = regime_classifier.compute_sampled_returns_with_regime_id(
        prices=prices,
        benchmark=benchmark,
        **regime_classifier.to_dict()
    )

    # Remove rows where all assets have missing returns
    sampled_returns_with_regime_id = sampled_returns_with_regime_id.dropna(how='all', axis=0)

    # Group returns by regime for regime-conditional analysis
    regime_groups = sampled_returns_with_regime_id.groupby([regime_classifier.REGIME_COLUMN], observed=False)

    # Build regression matrix with regime-specific benchmark returns as columns
    regression_datas = []
    for regime in regime_classifier.get_regime_ids():
        # Extract benchmark returns for this regime and rename column by regime name
        benchmark_data = regime_groups.get_group((regime,))[benchmark].rename(regime)
        regression_datas.append(benchmark_data)

    # Concatenate regime columns and fill missing values with zeros (non-regime periods)
    regression_matrix = pd.concat(regression_datas, axis=1).fillna(0)

    # Add constant term (alpha) to regression matrix if requested
    if is_add_alpha:
        #regression_matrix_with_const = sm.add_constant(regression_matrix).rename(
        #    columns={'const': ConditionalRegressionColumns.ALPHA_COLUMN.value}
        #)
        # Add separate alpha for each regime
        for regime in regime_classifier.get_regime_ids():
            regime_dummy = (sampled_returns_with_regime_id[regime_classifier.REGIME_COLUMN] == regime).astype(float)
            regression_matrix[f'{ConditionalRegressionColumns.ALPHA_COLUMN.value}_{regime}'] = regime_dummy
        regression_matrix_with_const = regression_matrix

    else:
        regression_matrix_with_const = regression_matrix

    # Prepare asset returns for regression (drop regime column and optionally benchmark)
    columns_to_drop = [regime_classifier.REGIME_COLUMN]
    if drop_benchmark:
        columns_to_drop.append(benchmark)
    asset_returns = sampled_returns_with_regime_id.drop(columns=columns_to_drop)

    # Create regime color mapping for visualization outputs
    regime_colors = regime_classifier.class_data_to_colors(
        sampled_returns_with_regime_id[regime_classifier.REGIME_COLUMN]
    ).rename(ConditionalRegressionColumns.COLOR_COLUMN.value)

    # Fit regression model for each asset
    model_params = []
    for asset in asset_returns.columns:
        y = asset_returns[asset]
        nan_mask = y.isna()

        # Check if all returns are missing
        if nan_mask.all():
            if len(model_params) == 0:
                # First asset must be benchmark - should never have all NaNs
                raise ValueError(f"Benchmark '{benchmark}' has all NaN returns - invalid specification")
            else:
                # For other assets, record NaN parameters matching structure of previous assets
                estimated_model_params = pd.Series(np.nan, index=model_params[0].index, name=asset)
        # Check if insufficient valid observations for regression
        elif (~nan_mask).sum() < min_period:
            estimated_model_params = pd.Series(np.nan, index=model_params[0].index, name=asset)
        else:
            # Filter to valid observations only
            if nan_mask.any():
                y_clean = y[~nan_mask]
            else:
                y_clean = y

            # Align regression matrix with valid y observations
            X = regression_matrix_with_const.loc[y_clean.index, :]

            # Fit OLS regression model
            model = sm.OLS(y_clean, X)
            estimated_model = model.fit()

            # Print regression summary if requested
            if is_print_summary:
                print(f"\n{'=' * 80}")
                print(f"Regression Summary for {asset}")
                print('=' * 80)
                print(estimated_model.summary())

            # Store detailed regression outputs if dict provided
            if regmodel_out_dict is not None:
                prediction = estimated_model.predict(regression_matrix_with_const)
                pandas_out = pd.concat([
                    sampled_returns_with_regime_id[regime_classifier.REGIME_COLUMN],
                    sampled_returns_with_regime_id[benchmark],
                    regression_matrix,
                    y,
                    prediction.rename(ConditionalRegressionColumns.PREDICTION.value),
                    regime_colors
                ], axis=1)
                regmodel_out_dict[asset] = pandas_out

            # Extract parameter estimates and residual variance
            estimated_model_params = estimated_model.params.copy()
            estimated_model_params[ConditionalRegressionColumns.RESID_VAR_COLUMN.value] = estimated_model.mse_resid
            estimated_model_params[ConditionalRegressionColumns.R2.value] = estimated_model.rsquared_adj
            estimated_model_params.name = asset

        # Append parameters for this asset
        model_params.append(estimated_model_params)

    # Combine all asset parameters into single DataFrame (assets as rows)
    model_params_df = pd.concat(model_params, axis=1).T

    return model_params_df


def get_regime_regression_params(prices: pd.DataFrame,
                                 regime_classifier: RegimeClassifier,
                                 benchmark: str,
                                 is_print_summary: bool = True,
                                 drop_benchmark: bool = False,
                                 min_period: int = 0,
                                 is_add_alpha: bool = False
                                 ) -> pd.DataFrame:
    """
    Convenience wrapper for estimate_cond_regression with common parameters.

    Args:
        prices: DataFrame of asset prices
        regime_classifier: Regime classification specifications
        benchmark: Benchmark asset name
        is_print_summary: Whether to print regression summaries
        drop_benchmark: Whether to exclude benchmark from results
        min_period: Minimum observations required per asset
        is_add_alpha: Whether to include regression intercept

    Returns:
        DataFrame of estimated regime betas per asset
    """
    estimated_params = estimate_cond_regression(
        prices=prices,
        benchmark=benchmark,
        regime_classifier=regime_classifier,
        is_print_summary=is_print_summary,
        drop_benchmark=drop_benchmark,
        min_period=min_period,
        is_add_alpha=is_add_alpha
    )
    return estimated_params


class LocalTests(Enum):
    """Enumeration of available local test cases."""
    REGRESSION = 1


def run_local_test(local_test: LocalTests):
    """
    Run local tests for development and debugging purposes.

    Integration tests that download real data and generate reports.
    Use for quick verification during development.

    Args:
        local_test: Which test case to run
    """
    from qis.test_data import load_etf_data

    # Load sample ETF price data
    prices = load_etf_data().dropna()
    print(prices)

    if local_test == LocalTests.REGRESSION:
        # Test regime-conditional regression with SPY as benchmark
        estimated_params = get_regime_regression_params(
            prices=prices,
            regime_classifier=BenchmarkReturnsQuantilesRegime(),
            benchmark='SPY',
            drop_benchmark=True,
            is_print_summary=True,
            is_add_alpha=True
        )
        print("\nEstimated Parameters:")
        print(estimated_params)


if __name__ == '__main__':
    run_local_test(local_test=LocalTests.REGRESSION)



"""
implementation of multi factor ewm model
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional, Union
from enum import Enum

# qis
import qis as qis
from qis import TimePeriod
import qis.utils.df_ops as dfo
import qis.perfstats.returns as ret
import qis.plots.time_series as pts
import qis.models.linear.ewm as ewm
from qis.models.linear.ewm import MeanAdjType, InitType
from qis.portfolio.risk.factor_model import LinearModel, compute_benchmarks_beta_attribution_from_prices


class EwmLinearModel(LinearModel):
    """Exponentially weighted moving average implementation of LinearModel."""

    def fit(self,
            span: Optional[int] = 31,
            ewm_lambda: float = 0.94,
            is_x_correlated: bool = True,
            mean_adj_type: MeanAdjType = MeanAdjType.NONE,
            init_type: InitType = InitType.MEAN,
            warmup_period: int = 20  # to avoid excessive betas at start,
            ) -> None:
        """Estimate time series EWM betas using exponential weighting.

        Args:
            span: Span for EWM calculation.
            ewm_lambda: Decay parameter for EWM.
            is_x_correlated: Whether to use diagonal (True) or full covariance matrix.
            mean_adj_type: Type of mean adjustment to apply.
            init_type: Initialization method for EWM.
        """
        x = self.x
        y = self.y
        if span is not None:
            ewm_lambda = 1.0 - 2.0 / (span + 1.0)
        if mean_adj_type != MeanAdjType.NONE:
            x = ewm.compute_rolling_mean_adj(data=x,
                                             mean_adj_type=mean_adj_type,
                                             ewm_lambda=ewm_lambda,
                                             init_type=init_type)

            y = ewm.compute_rolling_mean_adj(data=y,
                                             mean_adj_type=mean_adj_type,
                                             ewm_lambda=ewm_lambda,
                                             init_type=init_type)

        # compute list of betas using ewm numba recursion for cross product of x y and covariance of x
        # output is tensor of betas per date = [t, factors, assets]
        betas_ts = ewm.compute_ewm_xy_beta_tensor(x=x.to_numpy(),
                                                  y=y.to_numpy(),
                                                  ewm_lambda=ewm_lambda,
                                                  is_x_correlated=is_x_correlated,
                                                  warmup_period=warmup_period)
        # factor_loadings = {factor_id: pd.DataFrame(factor loadings)}
        loadings = dfo.np_txy_tensor_to_pd_dict(np_tensor_txy=betas_ts,
                                                dateindex=x.index,
                                                factor_names=x.columns.to_list(),
                                                asset_names=y.columns.to_list())
        self.x = x
        self.y = y
        self.loadings = loadings


def compute_portfolio_ewm_benchmark_betas(instrument_prices: pd.DataFrame,
                                          weights: pd.DataFrame,
                                          benchmark_prices: pd.DataFrame,
                                          time_period: TimePeriod = None,
                                          freq_beta: str = None,
                                          factor_beta_span: int = 63,  # quarter
                                          mean_adj_type: MeanAdjType = MeanAdjType.EWMA
                                          ) -> pd.DataFrame:
    """Compute portfolio benchmark betas using instrument exposures.

    Args:
        instrument_prices: Individual instrument price data.
        weights: Portfolio exposures to instruments.
        benchmark_prices: Benchmark price data.
        time_period: Optional time period filter.
        freq_beta: Frequency for return calculation.
        factor_beta_span: Span for EWM beta estimation.
        mean_adj_type: Mean adjustment method.

    Returns:
        Portfolio benchmark betas over time.
    """
    benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
    ewm_linear_model = EwmLinearModel(x=ret.to_returns(prices=benchmark_prices, freq=freq_beta, is_log_returns=True),
                                      y=ret.to_returns(prices=instrument_prices, freq=freq_beta, is_log_returns=True))
    ewm_linear_model.fit(span=factor_beta_span, is_x_correlated=True, mean_adj_type=mean_adj_type)
    weights = weights.reindex(index=instrument_prices.index, method='ffill')
    benchmark_betas = ewm_linear_model.compute_agg_factor_exposures(weights=weights)
    benchmark_betas = benchmark_betas.replace({0.0: np.nan}).ffill()  # fillholidays
    if time_period is not None:
        benchmark_betas = time_period.locate(benchmark_betas)
    return benchmark_betas


def compute_portfolio_benchmark_ewm_beta_alpha_attribution(instrument_prices: pd.DataFrame,
                                                           weights: pd.DataFrame,
                                                           benchmark_prices: pd.DataFrame,
                                                           portfolio_nav: pd.Series,
                                                           time_period: TimePeriod = None,
                                                           freq_beta: str = None,
                                                           factor_beta_span: int = 63,  # quarter
                                                           residual_name: str = 'Alpha'
                                                           ) -> pd.DataFrame:
    """Compute portfolio beta-alpha attribution using benchmark decomposition.

    Args:
        instrument_prices: Individual instrument prices.
        weights: Portfolio weights to instruments.
        benchmark_prices: Benchmark prices for attribution.
        portfolio_nav: Portfolio NAV time series.
        time_period: Optional time period filter.
        freq_beta: Frequency for beta estimation.
        factor_beta_span: EWM span for beta calculation.
        residual_name: Name for alpha/residual component.

    Returns:
        Attribution breakdown including alpha component.
    """
    portfolio_benchmark_betas = compute_portfolio_ewm_benchmark_betas(instrument_prices=instrument_prices,
                                                                      weights=weights,
                                                                      benchmark_prices=benchmark_prices,
                                                                      time_period=None,
                                                                      freq_beta=freq_beta,
                                                                      factor_beta_span=factor_beta_span)
    joint_attrib = compute_benchmarks_beta_attribution_from_prices(portfolio_nav=portfolio_nav,
                                                                   benchmark_prices=benchmark_prices,
                                                                   portfolio_benchmark_betas=portfolio_benchmark_betas,
                                                                   residual_name=residual_name,
                                                                   time_period=time_period)
    return joint_attrib


def estimate_ewm_factor_model(asset_prices: Union[pd.Series, pd.DataFrame],
                              factor_prices: Union[pd.Series, pd.DataFrame],
                              freq: str = 'W-WED',
                              span: int = 26,
                              mean_adj_type: MeanAdjType = MeanAdjType.NONE
                              ) -> EwmLinearModel:
    """Estimate linear factor model from price data.

    Args:
        asset_prices: Asset price time series.
        factor_prices: Factor price time series.
        freq: Frequency for return calculation.
        span: EWM span for model estimation.
        mean_adj_type: Mean adjustment method.

    Returns:
        Fitted EWM linear model.
    """
    y = qis.to_returns(asset_prices, freq=freq, is_log_returns=True, drop_first=True)
    x = qis.to_returns(factor_prices, freq=freq, is_log_returns=True, drop_first=True)
    ewm_linear_model = EwmLinearModel(x=x.reindex(index=y.index), y=y)
    ewm_linear_model.fit(span=span, is_x_correlated=True, mean_adj_type=mean_adj_type)
    return ewm_linear_model


class LocalTests(Enum):
    MODEL = 1
    ATTRIBUTION = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if local_test == LocalTests.MODEL:
        returns = np.log(prices.divide(prices.shift(1)))

        # factors
        factors = ['SPY', 'TLT', 'GLD']
        # factors = ['SPY']
        factor_returns = returns[factors]

        # assets
        is_check = False
        if is_check:
            asset_returns = returns[factors]
            asset_returns.columns = [f"{x.split('_')[0]}_asset" for x in factors]
        else:
            assets = ['QQQ', 'HYG']
            asset_returns = returns[assets]
        ewm_linear_model = EwmLinearModel(x=factor_returns, y=asset_returns)
        ewm_linear_model.fit(ewm_lambda=0.94, is_x_correlated=True)

        ewm_linear_model.print()
        ewm_linear_model.plot_factor_loadings(factor='SPY')

        factor_alpha, explained_returns = ewm_linear_model.get_factor_alpha()
        pts.plot_time_series(df=factor_alpha.cumsum(0), title='Cumulative alpha')
        pts.plot_time_series(df=explained_returns.cumsum(0), title='Cumulative explained return')

    elif local_test == LocalTests.ATTRIBUTION:
        benchmark_prices = prices[['SPY', 'TLT']]
        instrument_prices = prices[['QQQ', 'HYG', 'GLD']]
        exposures = pd.DataFrame(1.0/3.0, index=instrument_prices.index, columns=instrument_prices.columns)
        portfolio_nav = ret.returns_to_nav(returns=(exposures.shift(1)).multiply(instrument_prices.pct_change()).sum(axis=1))
        print(portfolio_nav)

        attribution = compute_portfolio_benchmark_ewm_beta_alpha_attribution(instrument_prices=instrument_prices,
                                                                             weights=exposures,
                                                                             benchmark_prices=benchmark_prices,
                                                                             portfolio_nav=portfolio_nav,
                                                                             time_period=None,
                                                                             freq_beta='W-WED',
                                                                             factor_beta_span=52,  # quarter
                                                                             residual_name='Alpha')
        pts.plot_time_series(df=attribution.cumsum(0))

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.MODEL)



# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum
import qis as qis
from qis import PerfParams, TimePeriod
import yfinance as yf


class LocalTests(Enum):
    LONG_IEF_SHORT_LQD = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    perf_params = PerfParams(freq_drawdown='B', freq='B')
    kwargs = dict(fontsize=12, digits_to_show=1, sharpe_digits=2,
                  alpha_format='{0:+0.0%}',
                  beta_format='{:0.1f}',
                  perf_stats_labels=qis.PerfStatsLabels.TOTAL_DETAILED.value,
                  framealpha=0.9)

    time_period = TimePeriod('31Dec2021', '19Sep2022')

    if local_test == LocalTests.LONG_IEF_SHORT_LQD:
        prices = yf.download(tickers=['IEF', 'LQD', 'LQDH', 'IGIB'], start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close']
        prices = time_period.locate(prices)
        rets = qis.to_returns(prices=prices, is_first_zero=True)
        navs1 = qis.returns_to_nav(returns=4.0*(rets.iloc[:, 0] - rets.iloc[:, 1]).rename('4x(Long 10y ETF / Short IG ETF)'))
        navs2 = qis.returns_to_nav(returns=-4.0*rets.iloc[:, 2].rename('4x(Short LQDH)'))
        navs3 = qis.returns_to_nav(returns=4.0*(0.5*rets.iloc[:, 0] - rets.iloc[:, 3]).rename('4x(Long 10y ETF / Short IGIB)'))
        spy = time_period.locate(yf.download(tickers=['SPY'], start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'])
        prices2 = pd.concat([navs1, navs2, navs3, spy], axis=1)

        with sns.axes_style('darkgrid'):
            fig1, axs = plt.subplots(2, 1, figsize=(15, 5), constrained_layout=True)
            qis.plot_prices_with_dd(prices=prices, perf_params=perf_params, axs=axs)

            fig2, ax = plt.subplots(1, 1, figsize=(12, 8), constrained_layout=True)
            qis.plot_prices(prices=prices2, perf_params=perf_params,
                            title='2022 YTD total performance of 4x(Long 10y ETF / Short IG ETF) vs SPY ETF',
                            ax=ax, **kwargs)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.LONG_IEF_SHORT_LQD)



"""
Regime classification for performance attribution analysis.

This module provides abstract and concrete implementations of regime classifiers
that partition time periods based on market conditions (e.g., bull/bear markets,
high/low volatility) for conditional performance analysis.
"""
from __future__ import annotations

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from abc import ABC, abstractmethod
from matplotlib._color_data import CSS4_COLORS as mcolors
from typing import Union, Dict, List, Tuple, Any
from enum import Enum

import qis.utils.df_cut as dfc
import qis.perfstats.perf_stats as pt
import qis.perfstats.returns as ret
from qis.perfstats.config import ReturnTypes, RegimeData, PerfParams, PerfStat
from qis.utils.annualisation import get_annualization_factor

# ============================================================================
# Core computation functions
# ============================================================================

def compute_mean_freq_regimes(sampled_returns_with_regime_id: pd.DataFrame) -> Tuple[pd.DataFrame, pd.Series]:
    """Compute mean returns and frequency for each regime.

    Args:
        sampled_returns_with_regime_id: DataFrame with returns and regime column

    Returns:
        Tuple of (regime means DataFrame, normalized regime frequencies Series)
    """
    regime_groups = sampled_returns_with_regime_id.groupby([RegimeClassifier.REGIME_COLUMN], observed=False)
    regime_means = regime_groups.mean()
    regime_dims = regime_groups.count().iloc[:, 0]

    # Replace nans and normalize
    regime_dims[np.isnan(regime_dims)] = 0.0
    norm_sum = np.sum(regime_dims)

    if np.isclose(norm_sum, 0.0):
        norm_q = np.zeros_like(regime_dims)
    else:
        norm_q = regime_dims / norm_sum

    return regime_means, norm_q


def compute_regime_avg(sampled_returns_with_regime_id: pd.DataFrame,
                       freq: str,
                       is_report_pa_returns: bool = True,
                       regime_ids: List[str] = None,
                       **kwargs
                       ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """Compute conditional means and annualized contributions by regime.

    Args:
        sampled_returns_with_regime_id: DataFrame with returns and regime classification
        freq: Sampling frequency for annualization
        is_report_pa_returns: If True, report as per annum returns
        regime_ids: Optional ordered list of regime IDs

    Returns:
        Tuple of (regime means, regime PA contributions, regime frequencies)
    """
    regime_means, norm_q = compute_mean_freq_regimes(
        sampled_returns_with_regime_id=sampled_returns_with_regime_id
    )

    af_mult = get_annualization_factor(freq=freq)

    if is_report_pa_returns:
        regime_pa = np.expm1(regime_means.multiply(af_mult * norm_q, axis=0))
    else:
        regime_pa = regime_means.multiply(af_mult * norm_q, axis=0)

    # Transpose: index = assets, columns = regimes
    regime_means = regime_means.T
    regime_pa = regime_pa.T

    # Arrange columns by specified regime order
    if regime_ids is not None:
        regime_means = regime_means[regime_ids]
        regime_pa = regime_pa[regime_ids]
    return regime_means, regime_pa, norm_q


def compute_regimes_pa_perf_table_from_sampled_returns(
        sampled_returns_with_regime_id: pd.DataFrame,
        prices: pd.DataFrame,
        benchmark: str,
        perf_params: PerfParams,
        freq: str,
        is_use_benchmark_means: bool = False,
        is_add_ra_perf_table: bool = True,
        drop_benchmark: bool = False,
        additive_pa_returns_to_pa_total: bool = True,
        regime_ids: List[str] = None,
        **kwargs) -> Tuple[pd.DataFrame, Dict[RegimeData, pd.DataFrame]]:
    """Compute comprehensive regime-conditional performance table.

    Args:
        sampled_returns_with_regime_id: Returns with regime classification
        prices: Asset price series
        benchmark: Benchmark asset name
        perf_params: Performance calculation parameters
        freq: Sampling frequency
        is_use_benchmark_means: Use benchmark means for normalization
        is_add_ra_perf_table: Include risk-adjusted performance table
        drop_benchmark: Exclude benchmark from final table
        additive_pa_returns_to_pa_total: Adjust regime PA to sum to total
        regime_ids: Ordered list of regime IDs

    Returns:
        Tuple of (performance table, regime data dictionary)
    """
    regime_avg, regime_pa, norm_q = compute_regime_avg(
        sampled_returns_with_regime_id=sampled_returns_with_regime_id,
        regime_ids=regime_ids,
        freq=freq,
        **kwargs
    )

    # Label columns consistently
    given_columns = regime_avg.columns.to_list()
    regime_avg = regime_avg[given_columns]
    regime_avg.columns = [f"{x} {RegimeData.REGIME_AVG.value}" for x in given_columns]

    regime_pa = regime_pa[given_columns]
    regime_pa.columns = [f"{x} {RegimeData.REGIME_PA.value}" for x in given_columns]
    regime_pa_columns = regime_pa.columns

    # Compute risk-adjusted performance table
    ra_perf_table = pt.compute_ra_perf_table_with_benchmark(
        prices=prices,
        benchmark=benchmark,
        perf_params=perf_params
    )

    if additive_pa_returns_to_pa_total:
        # Adjust regime PA to match total PA return
        total_sum = regime_pa[regime_pa_columns].sum(1)
        total_to_match = ra_perf_table[PerfStat.PA_RETURN.to_str()]
        total_pa_diff = total_to_match - total_sum

        weighted_diff = pd.DataFrame(
            np.tile(total_pa_diff, (len(norm_q.to_numpy()), 1)).T,
            index=total_pa_diff.index,
            columns=norm_q.index
        )
        regime_pa_diff = weighted_diff.multiply(norm_q, axis=1)
        regime_pa1 = regime_pa[regime_pa_columns].add(regime_pa_diff.to_numpy(), axis=0)
    else:
        regime_pa1 = regime_pa

    if is_use_benchmark_means and benchmark is not None:
        regime_pa1.loc[benchmark][regime_pa_columns] = regime_avg.loc[benchmark]

    # Compute regime Sharpe ratios
    vols_for_sharpe_pa = ra_perf_table[PerfStat.VOL.to_str()]
    regime_sharpe = regime_pa1.divide(vols_for_sharpe_pa, axis=0)[regime_pa_columns]
    regime_sharpe.columns = [f"{x}{RegimeData.REGIME_SHARPE.value}" for x in given_columns]

    # Combine into performance table
    if is_add_ra_perf_table:
        cond_perf_table = pd.concat([regime_avg, regime_pa1, regime_sharpe, ra_perf_table], axis=1)
    else:
        cond_perf_table = pd.concat([regime_avg, regime_pa1, regime_sharpe], axis=1)

    regime_datas = {
        RegimeData.REGIME_AVG: regime_avg,
        RegimeData.REGIME_PA: regime_pa1,
        RegimeData.REGIME_SHARPE: regime_sharpe
    }

    if drop_benchmark:
        cond_perf_table = cond_perf_table.drop(benchmark, axis=0)

    return cond_perf_table, regime_datas


# ============================================================================
# Abstract base class
# ============================================================================

class RegimeClassifier(ABC):
    """Abstract base class for regime classification.

    Regime classifiers partition time periods based on market conditions
    for conditional performance attribution analysis.
    """

    REGIME_COLUMN = 'regime'

    def __init__(self):
        self.regime_ids_colors: Dict[str, str] = {}
        super().__init__()

    @abstractmethod
    def compute_sampled_returns_with_regime_id(self, **kwargs) -> pd.DataFrame:
        """Compute returns with regime classification.

        Returns:
            DataFrame with returns and regime ID column
        """
        pass

    def get_regime_ids_colors(self) -> Dict[str, str]:
        """Get mapping of regime IDs to visualization colors.

        Returns:
            Dictionary mapping regime ID strings to color codes
        """
        return self.regime_ids_colors

    def get_regime_ids(self) -> List[str]:
        """Get ordered list of regime IDs.

        Returns:
            List of regime ID strings
        """
        return list(self.get_regime_ids_colors().keys())

    def to_dict(self) -> Dict[str, Any]:
        """Convert regime parameters to dictionary.

        Generic implementation that extracts public attributes (not starting with '_')
        from the instance, excluding methods and the REGIME_COLUMN class variable.
        Subclasses can override to customize serialization.

        Returns:
            Dictionary of parameter names to values

        Examples:
            >>> classifier = BenchmarkReturnsQuantilesRegime(freq='QE', q=4)
            >>> classifier.to_dict()
            {'freq': 'QE', 'return_type': <ReturnTypes.RELATIVE: ...>, 'q': 4, ...}
        """
        params = {}
        for key, value in self.__dict__.items():
            # Skip private attributes and methods
            if not key.startswith('_') and not callable(value):
                params[key] = value
        return params

    def compute_regimes_pa_perf_table(self,
                                      regime_id_func_kwargs: Dict[str, Any],
                                      prices: pd.DataFrame,
                                      benchmark: str,
                                      freq: str,
                                      perf_params: PerfParams,
                                      is_use_benchmark_means: bool = False,
                                      is_add_ra_perf_table: bool = True,
                                      drop_benchmark: bool = False,
                                      additive_pa_returns_to_pa_total: bool = True,
                                      regime_ids: List[str] = None,
                                      **kwargs) -> Tuple[pd.DataFrame, Dict[RegimeData, pd.DataFrame]]:
        """Compute regime-conditional performance attribution table.

        Args:
            regime_id_func_kwargs: Arguments for regime ID computation
            prices: Asset price series
            benchmark: Benchmark asset name
            freq: Sampling frequency
            perf_params: Performance parameters
            is_use_benchmark_means: Use benchmark means for normalization
            is_add_ra_perf_table: Include risk-adjusted performance
            drop_benchmark: Exclude benchmark from results
            additive_pa_returns_to_pa_total: Adjust regime PA to sum to total
            regime_ids: Ordered list of regime IDs

        Returns:
            Tuple of (performance table, regime data dictionary)
        """
        sampled_returns_with_regime_id = self.compute_sampled_returns_with_regime_id(
            **regime_id_func_kwargs
        )

        cond_perf_table, regime_datas = compute_regimes_pa_perf_table_from_sampled_returns(
            sampled_returns_with_regime_id=sampled_returns_with_regime_id,
            prices=prices,
            benchmark=benchmark,
            perf_params=perf_params,
            freq=freq,
            is_use_benchmark_means=is_use_benchmark_means,
            is_add_ra_perf_table=is_add_ra_perf_table,
            drop_benchmark=drop_benchmark,
            regime_ids=regime_ids
        )

        return cond_perf_table, regime_datas

    def class_data_to_colors(self, regime_data: pd.Series) -> pd.Series:
        """Map regime IDs to colors for visualization.

        Args:
            regime_data: Series of regime IDs

        Returns:
            Series of color codes
        """
        map_id_into_color = self.get_regime_ids_colors()
        regime_id_color = regime_data.map(map_id_into_color).astype(str)
        regime_id_color = regime_id_color.replace({'nan': '#FFFFFF'}) # type: ignore
        return regime_id_color


# ============================================================================
# Concrete implementations
# ============================================================================

class BenchmarkReturnsQuantilesRegime(RegimeClassifier):
    """Regime classifier based on benchmark return quantiles.

    Classifies periods into regimes based on quantiles of benchmark returns,
    enabling analysis of performance in different market environments
    (e.g., bear, normal, bull markets).
    """

    def __init__(self,
                 freq: str = 'QE',
                 return_type: ReturnTypes = ReturnTypes.RELATIVE,
                 q: Union[np.ndarray, int] = None,
                 regime_ids_colors: Dict[str, str] = None):
        """Initialize benchmark returns quantiles regime classifier.

        Args:
            freq: Sampling frequency (default: 'QE' for quarter-end)
            return_type: Type of returns to compute
            q: Quantile boundaries or number of quantiles (default: [0.0, 0.17, 0.83, 1.0])
            regime_ids_colors: Mapping of regime names to colors
        """
        super().__init__()
        self.freq = freq
        self.return_type = return_type
        self.q = q if q is not None else np.array([0.0, 0.17, 0.83, 1.0])
        self.regime_ids_colors = regime_ids_colors or {
            'Bear': mcolors['salmon'],
            'Normal': mcolors['yellowgreen'],
            'Bull': mcolors['darkgreen']
        }

    def compute_sampled_returns_with_regime_id(self,
                                               prices: Union[pd.DataFrame, pd.Series],
                                               benchmark: str,
                                               include_start_date: bool = True,
                                               include_end_date: bool = True,
                                               **kwargs) -> pd.DataFrame:
        """Classify periods by benchmark return quantiles.

        Args:
            prices: Asset prices
            benchmark: Benchmark column name
            include_start_date: Include first period
            include_end_date: Include last period

        Returns:
            DataFrame with returns and regime classification

        Raises:
            ValueError: If insufficient data for classification
        """
        if isinstance(prices, pd.Series):
            prices = prices.to_frame()

        sampled_returns_with_regime_id = ret.to_returns(
            prices=prices,
            freq=self.freq,
            return_type=self.return_type,
            include_start_date=include_start_date,
            include_end_date=include_end_date
        )

        if len(sampled_returns_with_regime_id.index) < 3:
            raise ValueError(
                f"Need at least 3 returns in time series: {sampled_returns_with_regime_id.index}\n"
                f"Decrease regime frequency"
            )

        x = sampled_returns_with_regime_id[benchmark]
        quant0 = pd.qcut(x=x, q=self.q, labels=self.get_regime_ids())
        sampled_returns_with_regime_id[self.REGIME_COLUMN] = quant0

        return sampled_returns_with_regime_id

    def compute_regimes_pa_perf_table(self,
                                      prices: pd.DataFrame,
                                      benchmark: str,
                                      perf_params: PerfParams,
                                      drop_benchmark: bool = False,
                                      **kwargs) -> Tuple[pd.DataFrame, Dict[RegimeData, pd.DataFrame]]:
        """Compute regime performance attribution table.

        Args:
            prices: Asset prices
            benchmark: Benchmark asset name
            perf_params: Performance parameters
            drop_benchmark: Exclude benchmark from results

        Returns:
            Tuple of (performance table, regime data dictionary)
        """
        regime_id_func_kwargs = dict(
            prices=prices,
            benchmark=benchmark,
            include_start_date=True,
            include_end_date=True
        )

        return super().compute_regimes_pa_perf_table(
            regime_id_func_kwargs=regime_id_func_kwargs,
            prices=prices,
            benchmark=benchmark,
            perf_params=perf_params,
            freq=self.freq,
            is_report_pa_returns=True,
            is_use_benchmark_means=False,
            regime_ids=self.get_regime_ids(),
            drop_benchmark=drop_benchmark
        )


class BenchmarkReturnsPositiveNegativeRegime(RegimeClassifier):
    """Regime classifier based on positive vs negative benchmark returns.

    Classifies periods into two simple regimes based on benchmark return sign,
    useful for up/down market analysis.
    """

    def __init__(self,
                 freq: str = 'QE',
                 return_type: ReturnTypes = ReturnTypes.RELATIVE,
                 regime_ids_colors: Dict[str, str] = None):
        """Initialize positive/negative regime classifier.

        Args:
            freq: Sampling frequency (default: 'QE' for quarter-end)
            return_type: Type of returns to compute
            regime_ids_colors: Mapping of regime names to colors
        """
        super().__init__()
        self.freq = freq
        self.return_type = return_type
        self.regime_ids_colors = regime_ids_colors or {
            'Negative': mcolors['salmon'],
            'Positive': mcolors['darkgreen']
        }

    def compute_sampled_returns_with_regime_id(self,
                                               prices: Union[pd.DataFrame, pd.Series],
                                               benchmark: str,
                                               include_start_date: bool = True,
                                               include_end_date: bool = True,
                                               **kwargs) -> pd.DataFrame:
        """Classify periods by benchmark return sign.

        Args:
            prices: Asset prices
            benchmark: Benchmark column name
            include_start_date: Include first period
            include_end_date: Include last period

        Returns:
            DataFrame with returns and regime classification

        Raises:
            ValueError: If insufficient data for classification
        """
        if isinstance(prices, pd.Series):
            prices = prices.to_frame()

        sampled_returns_with_regime_id = ret.to_returns(
            prices=prices,
            freq=self.freq,
            return_type=self.return_type,
            include_start_date=include_start_date,
            include_end_date=include_end_date
        )

        if len(sampled_returns_with_regime_id.index) < 2:
            raise ValueError(
                f"Need at least 2 returns in time series: {sampled_returns_with_regime_id.index}\n"
                f"Decrease regime frequency"
            )

        benchmark_returns = sampled_returns_with_regime_id[benchmark]
        regime_ids = self.get_regime_ids()

        regime_classification = pd.Series(
            np.where(benchmark_returns < 0, regime_ids[0], regime_ids[1]),
            index=benchmark_returns.index,
            name=self.REGIME_COLUMN
        )

        sampled_returns_with_regime_id[self.REGIME_COLUMN] = regime_classification

        return sampled_returns_with_regime_id

    def compute_regimes_pa_perf_table(self,
                                      prices: pd.DataFrame,
                                      benchmark: str,
                                      perf_params: PerfParams,
                                      drop_benchmark: bool = False,
                                      **kwargs) -> Tuple[pd.DataFrame, Dict[RegimeData, pd.DataFrame]]:
        """Compute regime performance attribution table.

        Args:
            prices: Asset prices
            benchmark: Benchmark asset name
            perf_params: Performance parameters
            drop_benchmark: Exclude benchmark from results

        Returns:
            Tuple of (performance table, regime data dictionary)
        """
        regime_id_func_kwargs = dict(
            prices=prices,
            benchmark=benchmark,
            include_start_date=True,
            include_end_date=True
        )

        return super().compute_regimes_pa_perf_table(
            regime_id_func_kwargs=regime_id_func_kwargs,
            prices=prices,
            benchmark=benchmark,
            perf_params=perf_params,
            freq=self.freq,
            is_report_pa_returns=True,
            is_use_benchmark_means=False,
            regime_ids=self.get_regime_ids(),
            drop_benchmark=drop_benchmark
        )


class BenchmarkVolsQuantilesRegime(RegimeClassifier):
    """Regime classifier based on benchmark volatility quantiles.

    Classifies periods based on realized volatility levels, enabling analysis
    of performance in different volatility environments.
    """

    def __init__(self,
                 freq: str = 'QE',
                 return_type: ReturnTypes = ReturnTypes.RELATIVE,
                 q: int = 4):
        """Initialize volatility quantiles regime classifier.

        Args:
            freq: Sampling frequency for volatility calculation
            return_type: Type of returns to compute
            q: Number of quantile buckets
        """
        super().__init__()
        self.freq = freq
        self.return_type = return_type
        self.q = q
        self.regime_colors: Dict[str, Tuple[float, ...]] = {}

    def compute_sampled_returns_with_regime_id(self,
                                               prices: pd.DataFrame,
                                               benchmark: str,
                                               include_start_date: bool = True,
                                               include_end_date: bool = True,
                                               **kwargs) -> pd.DataFrame:
        """Classify periods by benchmark volatility quantiles.

        Args:
            prices: Asset prices
            benchmark: Benchmark column name
            include_start_date: Include first period
            include_end_date: Include last period

        Returns:
            DataFrame with returns and regime classification
        """
        vols = ret.compute_sampled_vols(
            prices=prices[benchmark],
            freq_vol=self.freq,
            include_start_date=include_start_date,
            include_end_date=include_end_date
        )

        hue_name = f"{benchmark} vol"
        classificator, labels = dfc.add_quantile_classification(
            df=vols.to_frame(),
            x_column=benchmark,
            num_buckets=self.q,
            hue_name=hue_name,
            xvar_format='{:.0%}',
            bucket_prefix=hue_name
        )

        classificator = classificator.sort_index()

        sampled_returns_with_regime_id = ret.to_returns(
            prices=prices,
            freq=self.freq,
            return_type=self.return_type,
            include_start_date=include_start_date,
            include_end_date=include_end_date
        )

        sampled_returns_with_regime_id[self.REGIME_COLUMN] = classificator[hue_name]

        # Generate colors using colormap
        cmap = plt.get_cmap('RdYlGn', len(labels))
        colors = [cmap(n) for n in range(len(labels))]
        self.regime_colors = {k: v for k, v in zip(labels, colors)}

        return sampled_returns_with_regime_id

    def compute_regimes_pa_perf_table(self,
                                      prices: pd.DataFrame,
                                      benchmark: str,
                                      perf_params: PerfParams,
                                      **kwargs) -> Tuple[pd.DataFrame, Dict[RegimeData, pd.DataFrame]]:
        """Compute regime performance attribution table.

        Args:
            prices: Asset prices
            benchmark: Benchmark asset name
            perf_params: Performance parameters

        Returns:
            Tuple of (performance table, regime data dictionary)
        """
        regime_id_func_kwargs = dict(
            prices=prices,
            benchmark=benchmark,
            include_start_date=True,
            include_end_date=True
        )

        return super().compute_regimes_pa_perf_table(
            regime_id_func_kwargs=regime_id_func_kwargs,
            prices=prices,
            benchmark=benchmark,
            perf_params=perf_params,
            freq=self.freq,
            is_report_pa_returns=True,
            is_use_benchmark_means=False,
            regime_ids=self.get_regime_ids()
        )

    def get_regime_colors(self) -> List[Tuple[float, ...]]:
        """Get list of regime colors in order."""
        return list(self.regime_colors.values())


# ============================================================================
# Convenience functions
# ============================================================================

def compute_bnb_regimes_pa_perf_table(prices: pd.DataFrame,
                                      benchmark: str = None,
                                      benchmark_price: pd.Series = None,
                                      freq: str = 'QE',
                                      return_type: ReturnTypes = ReturnTypes.RELATIVE,
                                      q: Union[np.ndarray, int] = None,
                                      regime_ids_colors: Dict[str, str] = None,
                                      perf_params: PerfParams = None,
                                      drop_benchmark: bool = False,
                                      **kwargs) -> pd.DataFrame:
    """Compute benchmark regime performance attribution table.

    Convenience function for computing regime-conditional performance using
    benchmark return quantiles classification.

    Args:
        prices: Asset price series
        benchmark: Benchmark column name in prices
        benchmark_price: Alternative benchmark price series to add to prices
        freq: Sampling frequency
        return_type: Type of returns to compute
        q: Quantile boundaries or number of quantiles
        regime_ids_colors: Mapping of regime names to colors
        perf_params: Performance parameters
        drop_benchmark: Exclude benchmark from results

    Returns:
        Regime-conditional performance table

    Raises:
        ValueError: If neither benchmark nor benchmark_price provided
    """
    if benchmark is None and benchmark_price is None:
        raise ValueError("Provide either benchmark name in prices or benchmark_price")

    if benchmark is not None and benchmark_price is None:
        if benchmark not in prices.columns:
            raise ValueError(f"{benchmark} is not in {prices.columns.to_list()}")
    elif benchmark_price is not None:
        if benchmark not in prices.columns:
            if not isinstance(benchmark_price, pd.Series):
                raise ValueError(f"benchmark_price must be pd.Series not {type(benchmark_price)}")
            benchmark_price = benchmark_price.reindex(index=prices.index, method='ffill').ffill()
            prices = pd.concat([benchmark_price, prices], axis=1)
            benchmark = benchmark_price.name

    regime_classifier = BenchmarkReturnsQuantilesRegime(
        freq=freq,
        return_type=return_type,
        q=q,
        regime_ids_colors=regime_ids_colors
    )

    regimes_pa_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(
        prices=prices,
        benchmark=benchmark,
        perf_params=perf_params,
        drop_benchmark=drop_benchmark
    )

    return regimes_pa_perf_table


class LocalTests(Enum):
    """Test cases for local development."""
    BNB_REGIME = 1
    BNB_PERF_TABLE = 2
    POS_NEG_REGIME = 3
    VOL_REGIME = 4
    TO_DICT = 5


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging.

    Integration tests that download real data and generate reports
    for quick verification during development.

    Args:
        local_test: Test case to run
    """
    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()
    perf_params = PerfParams()

    if local_test == LocalTests.BNB_REGIME:
        regime_classifier = BenchmarkReturnsQuantilesRegime(
            freq='QE',
            q=np.array([0.0, 0.17, 0.83, 1.0])
        )

        regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(
            prices=prices,
            benchmark='SPY'
        )
        print(f"regime_ids:\n{regime_ids}")

        cond_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(
            prices=prices,
            benchmark='SPY',
            perf_params=perf_params
        )
        print(f"\nregime_means:\n{cond_perf_table}")
        print(f"\nregime_pa:\n{regime_datas}")

    elif local_test == LocalTests.BNB_PERF_TABLE:
        df = compute_bnb_regimes_pa_perf_table(
            prices=prices,
            benchmark='SPY',
            perf_params=PerfParams()
        )
        print(df)
        print(df.columns)

    elif local_test == LocalTests.POS_NEG_REGIME:
        regime_classifier = BenchmarkReturnsPositiveNegativeRegime(freq='QE')

        regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(
            prices=prices,
            benchmark='SPY'
        )
        print(f"regime_ids:\n{regime_ids}")

        cond_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(
            prices=prices,
            benchmark='SPY',
            perf_params=perf_params
        )
        print(f"\ncond_perf_table:\n{cond_perf_table}")

    elif local_test == LocalTests.VOL_REGIME:
        regime_classifier = BenchmarkVolsQuantilesRegime(freq='QE', q=4)

        regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(
            prices=prices,
            benchmark='SPY'
        )
        print(f"regime_ids:\n{regime_ids}")

        cond_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(
            prices=prices,
            benchmark='SPY',
            perf_params=perf_params
        )
        print(f"\ncond_perf_table:\n{cond_perf_table}")

    elif local_test == LocalTests.TO_DICT:
        print("Testing to_dict() method for all regime classifiers\n")

        # Test BenchmarkReturnsQuantilesRegime
        classifier1 = BenchmarkReturnsQuantilesRegime(
            freq='QE',
            return_type=ReturnTypes.RELATIVE,
            q=np.array([0.0, 0.25, 0.75, 1.0])
        )
        print("BenchmarkReturnsQuantilesRegime.to_dict():")
        print(classifier1.to_dict())
        print()

    plt.show()


if __name__ == '__main__':
    run_local_test(local_test=LocalTests.TO_DICT)


"""
analytic for melting of pandas
"""
# packages
import pandas as pd
import numpy as np
from enum import Enum
from typing import Dict, Optional, List
# qis
import qis.utils.np_ops as npo


def melt_scatter_data_with_xvar(df: pd.DataFrame,
                                xvar_str: str,
                                y_column: str = 'Strategy returns',
                                hue_name: str = 'hue'
                                ) -> pd.DataFrame:
    """
    column with xvar_str will be repeated to melted y_column
    index is ignored
    """
    df = df.dropna()
    ex_benchmark_data = df.drop(xvar_str, axis=1)
    scatter_data = pd.melt(df,
                           value_vars=ex_benchmark_data.columns,
                           id_vars=[xvar_str],
                           var_name=hue_name,
                           value_name=y_column)
    # move hue to last position
    columns_ex_hue = scatter_data.columns.to_list()
    columns_ex_hue.remove(hue_name)
    scatter_data = scatter_data[columns_ex_hue+[hue_name]]
    return scatter_data


def melt_scatter_data_with_xdata(df: pd.DataFrame,
                                 xdata: pd.Series,
                                 y_column: str = 'Vars',
                                 hue_name: str = 'hue'
                                 ) -> pd.DataFrame:
    """
    df will be melted using xdata
    must be same indices
    """

    joint_data = pd.concat([xdata, df], axis=1)
    scatter_data = pd.melt(joint_data,
                           value_vars=df.columns,
                           id_vars=[xdata.name],
                           var_name=hue_name,
                           value_name=y_column)

    # move hue to last position
    columns_ex_hue = scatter_data.columns.to_list()
    columns_ex_hue.remove(hue_name)
    scatter_data = scatter_data[columns_ex_hue+[hue_name]]

    return scatter_data


def melt_df_by_columns(df: pd.DataFrame,
                       x_index_var_name: Optional[str] = 'date',
                       y_var_name: str = 'returns',
                       hue_var_name: str = 'instrument',
                       hue_order: List[str] = None,
                       ) -> pd.DataFrame:
    """
    index is added to hue variables
    df melted to hue = [index_values, column_values, data_values]
    -> df.columns = [x_index_var_name, hue_var_name, y_var_name]
    """
    df = df.dropna(axis=1, how='all')
    df.index.name = x_index_var_name # set to match id_vars
    box_data = pd.melt(df.reset_index(),
                       id_vars=x_index_var_name,
                       value_vars=df.columns.to_list(),
                       value_name=y_var_name,
                       var_name=hue_var_name)

    if hue_order is not None:  # sort by hue
        sort_column = 'sort_column'
        name_sort = {key: idx for idx, key in enumerate(hue_order)}
        box_data[sort_column] = box_data[x_index_var_name].map(name_sort)
        box_data = box_data.sort_values(by=sort_column).drop(sort_column, axis=1)

    box_data = box_data.loc[pd.isna(box_data[y_var_name]) == False, :]  # important to exclude nans

    return box_data


def melt_paired_df(indicator: pd.DataFrame,
                   observations: pd.DataFrame,
                   signal_name: str = 'signal',
                   ra_return_name: str = 'ra_return',
                   hue_name: str = 'hue'
                   ) -> pd.DataFrame:
    # melt to pandas
    col_union = observations.columns.intersection(indicator.columns)
    # indicator = indicator[col_union]
    observations = observations[indicator.columns]
    temp_hue = f"{hue_name}_y"
    x_pd = pd.melt(indicator, value_vars=indicator.columns.to_list(), var_name=hue_name, value_name=signal_name)
    y_pd = pd.melt(observations, value_vars=observations.columns.to_list(), var_name=temp_hue, value_name=ra_return_name)
    # concat
    scatter_data = pd.concat([x_pd, y_pd], axis=1).dropna()
    scatter_data = scatter_data.drop(columns=[temp_hue])
    scatter_data = scatter_data.sort_values(by=signal_name)
    return scatter_data


class SignCondition(Enum):
    NONE = 'None'
    TREND = 'Trend'
    REVERSION = 'Reversion'


def melt_signed_paired_df(observations: pd.DataFrame,
                          indicator: pd.DataFrame,
                          signal_name: str = 'signal',
                          ra_return_name: str = 'ra_return',
                          hue_name: str = 'hue'
                          ) -> Dict[SignCondition, pd.DataFrame]:
    # melt to pandas
    scatter_data = melt_paired_df(indicator=indicator,
                                  observations=observations,
                                  signal_name=signal_name,
                                  ra_return_name=ra_return_name,
                                  hue_name=hue_name)

    joint_cond, trend_cond, rev_cond = npo.compute_paired_signs(x=scatter_data[signal_name].to_numpy(),
                                                                y=scatter_data[ra_return_name].to_numpy())

    data_out = {SignCondition.NONE: scatter_data.loc[joint_cond, :],
                SignCondition.TREND: scatter_data.loc[trend_cond, :],
                SignCondition.REVERSION: scatter_data.loc[rev_cond, :]}

    return data_out


class LocalTests(Enum):
    PD_MELT = 1
    SCATTER_DATA = 2
    MELT_DF_BY_COLUMNS = 3


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """
    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna().asfreq('QE', method='ffill')
    returns = prices.pct_change()

    if local_test == LocalTests.PD_MELT:
        df = pd.DataFrame(data=[[0, 1, 2.5, 0], [2, 3, 5.0, 0], [np.nan, 1, 1, 1]],
                          index=['cat', 'dog', 'missing'],
                          columns=['weight', 'height', 'age', 'sex'])
        print(df)
        melted = pd.melt(df, value_vars=df.columns, var_name='myVarname', value_name='myValname')
        print(f"melted=\n{melted}")

        scatter_data = melt_scatter_data_with_xvar(df=df, xvar_str='age', y_column='weight_height')
        print(f"scatter_data=\n{scatter_data}")

        box_data = melt_df_by_columns(df=df, x_index_var_name='animal', hue_var_name='hue_features', y_var_name='observations')
        print(f"box_data=\n{box_data}")

    elif local_test == LocalTests.SCATTER_DATA:
        scatter_data = melt_scatter_data_with_xvar(df=returns, xvar_str='SPY')
        print(scatter_data)

    elif local_test == LocalTests.MELT_DF_BY_COLUMNS:
        box_data = melt_df_by_columns(df=returns)
        print(box_data)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.PD_MELT)
